import io
import json
from typing import Any

from aqueduct_executor.operators.utils.storage.config import GCSStorageConfig
from aqueduct_executor.operators.utils.storage.storage import Storage
from google.cloud import storage
from google.oauth2 import service_account


class GCSStorage(Storage):
    _client: Any  # GCS client
    _config: GCSStorageConfig

    def __init__(self, config: GCSStorageConfig):
        credentials_info = json.loads(config.service_account_credentials)
        credentials = service_account.Credentials.from_service_account_info(credentials_info)
        self._client = storage.Client(credentials=credentials)
        self._config = config

    def put(self, key: str, value: bytes) -> None:
        bucket = self._client.bucket(self._config.bucket)
        blob = bucket.blob(key)

        print(f"writing to gcs: {key}")
        f = io.BytesIO(value)
        blob.upload_from_file(f)

    def get(self, key: str) -> bytes:
        bucket = self._client.bucket(self._config.bucket)
        blob = bucket.blob(key)

        print(f"reading from gcs: {key}")
        return bytes(blob.download_as_bytes())



import sys

from aqueduct_executor.operators.system_metric_executor.spec import SystemMetricSpec
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import ArtifactType, ExecutionStatus, FailureType
from aqueduct_executor.operators.utils.execution import (
    TIP_UNKNOWN_ERROR,
    Error,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.storage.parse import parse_storage


def run(spec: SystemMetricSpec) -> None:
    """
    Executes a system metric operator by storing the requested system metrics value in the output content path.
    """
    print("Job Spec: \n{}".format(spec.json()))

    storage = parse_storage(spec.storage_config)
    exec_state = ExecutionState(user_logs=Logs())
    try:
        # We currently allow the spec to contain multiple input_metadata paths.
        # A system metric currently spans over a single operator.
        # The scheduler enforces this requirement before the executor is run.
        system_metadata = utils.read_system_metadata(storage, spec.input_metadata_paths)
        utils.write_artifact(
            storage,
            ArtifactType.NUMERIC,
            False,  # derived_from_bson
            spec.output_content_path,
            spec.output_metadata_path,
            float(system_metadata[0][spec.metric_name]),
            system_metadata={},
        )
        exec_state.status = ExecutionStatus.SUCCEEDED
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
    except ExecFailureException as e:
        from_exception_exec_state = ExecutionState.from_exception(e, user_logs=Logs())
        print(f"Failed with error. Full Logs:\n{from_exception_exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, from_exception_exec_state)
        sys.exit(1)
    except Exception as e:
        exec_state.status = ExecutionStatus.FAILED
        exec_state.failure_type = FailureType.SYSTEM
        exec_state.error = Error(context=exception_traceback(e), tip=TIP_UNKNOWN_ERROR)
        print(f"Failed with system error. Full Logs:\n{exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)



"""
CONFIG FILE FOR TABLE CONNECTOR INTEGRATION TESTS
- To skip a particular test set the relevant `SKIP_` flag to False
- Set the `_CONF` dict for all connectors being tested.
- Commented out config fields are optional.
"""
from aqueduct_executor.operators.connectors.data import (
    bigquery,
    mysql,
    postgres,
    snowflake,
    sql_server,
    sqlite,
)
from aqueduct_executor.operators.connectors.data.config import BigQueryConfig

"""FLAGS TO SKIP TESTS"""
SKIP_POSTGRES = True
SKIP_SNOWFLAKE = True
SKIP_MYSQL = True
SKIP_REDSHIFT = True
SKIP_MARIADB = True
SKIP_SQL_SERVER = True
SKIP_BIGQUERY = True
SKIP_SQLITE = True

# """POSTGRES CONFIG"""
# POSTGRES_CONF = {
#     postgres._CONFIG_USERNAME_KEY: "",
#     postgres._CONFIG_PASSWORD_KEY: "",
#     postgres._CONFIG_DATABASE_KEY: "",
#     postgres._CONFIG_HOST_KEY: "",
#     # postgres._CONFIG_PORT_KEY: "",
# }

# """SNOWFLAKE CONFIG"""
# SNOWFLAKE_CONF = {
#     snowflake._CONFIG_USERNAME_KEY: "",
#     snowflake._CONFIG_PASSWORD_KEY: "",
#     snowflake._CONFIG_ACCOUNT_IDENTIFIER_KEY: "",
#     snowflake._CONFIG_DATABASE_KEY: "",
#     snowflake._CONFIG_WAREHOUSE_KEY: "",
#     # snowflake._CONFIG_SCHEMA_KEY: "",
# }

# """MYSQL CONFIG"""
# MYSQL_CONF = {
#     mysql._CONFIG_USERNAME_KEY: "",
#     mysql._CONFIG_PASSWORD_KEY: "",
#     mysql._CONFIG_DATABASE_KEY: "",
#     mysql._CONFIG_HOST_KEY: "",
#     mysql._CONFIG_PORT_KEY: "",
# }

# """REDSHIFT CONFIG"""
# REDSHIFT_CONF = {
#     postgres._CONFIG_USERNAME_KEY: "",
#     postgres._CONFIG_PASSWORD_KEY: "",
#     postgres._CONFIG_DATABASE_KEY: "",
#     postgres._CONFIG_HOST_KEY: "",
#     postgres._CONFIG_PORT_KEY: "",
# }

# """MARIADB CONFIG"""
# MARIADB_CONF = {
#     mysql._CONFIG_USERNAME_KEY: "",
#     mysql._CONFIG_PASSWORD_KEY: "",
#     mysql._CONFIG_DATABASE_KEY: "",
#     mysql._CONFIG_HOST_KEY: "",
#     mysql._CONFIG_PORT_KEY: "",
# }

# """SQL SERVER CONFIG"""
# SQL_SERVER_CONF = {
#     sql_server._CONFIG_USERNAME_KEY: "",
#     sql_server._CONFIG_PASSWORD_KEY: "",
#     sql_server._CONFIG_DATABASE_KEY: "",
#     sql_server._CONFIG_HOST_KEY: "",
#     sql_server._CONFIG_PORT_KEY: "",
# }

"""BIGQUERY CONFIG"""
BIGQUERY_CONF = BigQueryConfig(project_id="", service_account_credentials="")

# """SQLITE CONFIG"""
# SQLITE_CONF = {
#     sqlite._CONFIG_DATABASE_PATH_KEY: "",
# }



from typing import Dict, List, Union

try:
    from typing import Literal
except ImportError:
    # Python 3.7 does not support typing.Literal
    from typing_extensions import Literal  # type: ignore

import json

from aqueduct_executor.operators.connectors.data import common, config, extract, load, models
from aqueduct_executor.operators.utils import enums
from aqueduct_executor.operators.utils.storage import config as sconfig
from pydantic import parse_obj_as, validator

AQUEDUCT_DEMO_NAME = "aqueduct_demo"


def unwrap_connector_config(cls, connector_config, values):  # type: ignore
    """
    TODO ENG-937: Remove this validator once connector config serialization is fixed.

    Unwraps the connector config before it can be parsed into a
    config.Config object. This is necessary because of how connector_config
    is serialized in Golang.

    For non-OAuth configs, it has the following structure:
    "connector_config": {
        "conf": {
            "username": "username",
            "password": "password",
        }
    }

    For OAuth configs, it has the following structure:
    "connector_config": {
        "token": {
            "access_token": "123456",
            "refresh_token": "123",
        },
        "oauth2_conf": {...},
        "public_conf": {...},
    }
    """

    if "connector_name" not in values:
        raise ValueError("Unknown connector name.")

    if not isinstance(connector_config, dict):
        raise ValueError("connector_config is not a dictionary.")

    if type(values["connector_name"]) == dict:
        for integration in connector_config:
            if "conf" in connector_config[integration]:
                connector_config[integration] = connector_config[integration]["conf"]

        return connector_config
    else:
        if "conf" not in connector_config:
            # There is no inner `conf` dictionary to unwrap
            # This occurs when the spec is serialized in Python to run operators
            # on other engines.
            return connector_config

        # This is a static config
        return connector_config["conf"]


class AuthenticateSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.AUTHENTICATE]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class ExtractSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.EXTRACT]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    parameters: extract.Params

    # The input fields are only used to record user-defined parameters for relational queries.
    # The tags in the queries will be expanded into the parameter values.
    input_param_names: List[str]
    input_content_paths: List[str]
    input_metadata_paths: List[str]  # This field is ignored and is only here for completeness.
    output_content_path: str
    output_metadata_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class LoadSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.LOAD]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    parameters: load.Params
    input_content_path: str
    input_metadata_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class LoadTableSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.LOADTABLE]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    csv: str
    load_parameters: LoadSpec

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class DeleteSavedObjectsSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.DELETESAVEDOBJECTS]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: Dict[str, common.Name]
    connector_config: Dict[str, config.Config]
    integration_to_object: Dict[str, List[str]]
    output_content_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class DiscoverSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.DISCOVER]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    output_content_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


Spec = Union[
    AuthenticateSpec, ExtractSpec, LoadSpec, LoadTableSpec, DiscoverSpec, DeleteSavedObjectsSpec
]


def parse_spec(spec_json: bytes) -> Spec:
    """
    Parses a JSON string into a Spec.
    """
    data = json.loads(spec_json)

    # TODO (ENG-1286): https://linear.app/aqueducthq/issue/ENG-1286/investigate-why-mypy-is-complaining-about-object-parsing
    # The following line is working, but mypy complains:
    # Argument 1 to "parse_obj_as" has incompatible type "object"; expected "Type[<nothing>]"
    # We ignore the error for now.
    return parse_obj_as(Spec, data)  # type: ignore



import pytest
from aqueduct_executor.operators.connectors.data import bigquery as bq
from aqueduct_executor.operators.connectors.data.extract import RelationalParams as ExtractParam
from aqueduct_executor.operators.connectors.data.load import RelationalParams as LoadParam
from aqueduct_executor.operators.connectors.tests import conf, utils
from google.cloud import bigquery

_DATASET = "testdata"
_TABLE = "testdata.bigquery"


@pytest.mark.skipif(conf.SKIP_BIGQUERY, reason="Skip BigQuery Flag Set")
class TestBigQuery:
    @classmethod
    def setup_class(cls):
        # Setup connector
        config = conf.BIGQUERY_CONF
        conn = bq.BigQueryConnector(config)
        cls.conn = conn

        # Setup test dataframe
        cls.test_df = utils.sample_df()

        # Create BigQuery client to perform setup and teardown
        cls.client = bigquery.Client(project=conn.project_id, credentials=conn.credentials)
        cls.client.create_dataset(_DATASET)

    @classmethod
    def teardown_class(cls):
        cls.client.delete_dataset(_DATASET, delete_contents=True)

    def test_authenticate(self):
        utils.authenticate_test(self.conn)

    @pytest.mark.dependency()
    def test_load(self):
        params = LoadParam(table=_TABLE)
        utils.load_test(self.conn, params, self.test_df)

    @pytest.mark.dependency(depends=["TestBigQuery::test_load"])
    def test_extract(self):
        params = ExtractParam(query="SELECT * FROM {};".format(_TABLE))
        utils.extract_test(self.conn, params, expected_df=self.test_df)

    def test_discover(self):
        tables = set()
        for i in range(2):
            table = f"{_DATASET}.table{i}"
            params = LoadParam(table=table)
            utils.load_test(self.conn, params, self.test_df)
            tables.add(table)
            discover = self.conn.discover()
            assert len(tables.difference(discover)) == 0



import json
from typing import Any, List, Optional

try:
    from typing import Literal
except ImportError:
    # Python 3.7 does not support typing.Literal
    from typing_extensions import Literal  # type: ignore

from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    CheckSeverity,
    JobType,
    OperatorType,
)
from aqueduct_executor.operators.utils.storage import config
from pydantic import BaseModel, Extra, parse_obj_as


class FunctionSpec(BaseModel):
    name: str
    type: Literal[JobType.FUNCTION]
    storage_config: config.StorageConfig
    metadata_path: str
    function_path: str
    function_extract_path: str
    entry_point_file: str
    entry_point_class: str
    entry_point_method: str
    custom_args: str
    input_content_paths: List[str]
    input_metadata_paths: List[str]
    output_content_paths: List[str]
    output_metadata_paths: List[str]
    expected_output_artifact_types: List[ArtifactType]
    operator_type: OperatorType

    # This is specific to the check operator. This is left unset by any other function type.
    check_severity: Optional[CheckSeverity]

    # This is always unset - it is only here because we forbid extra fields.
    resources: Optional[Any]

    class Config:
        extra = Extra.forbid


def parse_spec(spec_json: bytes) -> FunctionSpec:
    """
    Parses a JSON string into a FunctionSpec.
    """
    data = json.loads(spec_json)
    return parse_obj_as(FunctionSpec, data)



import io
import json
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
from aqueduct.utils.serialization import (
    DEFAULT_ENCODING,
    artifact_type_to_serialization_type,
    deserialize,
    serialize_val,
)
from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    ExecutionStatus,
    FailureType,
    SerializationType,
)
from aqueduct_executor.operators.utils.exceptions import MissingInputPathsException
from aqueduct_executor.operators.utils.execution import (
    TIP_UNKNOWN_ERROR,
    Error,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.storage.storage import Storage

_RUNTIME_SEC_METRIC_NAME = "runtime"
_MAX_MEMORY_MB_METRIC_NAME = "max_memory"
_METADATA_SCHEMA_KEY = "schema"
_METADATA_SYSTEM_METADATA_KEY = "system_metadata"
_METADATA_ARTIFACT_TYPE_KEY = "artifact_type"
_METADATA_SERIALIZATION_TYPE_KEY = "serialization_type"
_METADATA_PYTHON_TYPE_KEY = "python_type"

# The temporary file name that a Tensorflow keras model will be dumped into before we read/write it from storage.
# This will be cleaned up within the serialization logic.
_TEMP_KERAS_MODEL_NAME = "keras_model"


def _read_csv(input_bytes: bytes) -> pd.DataFrame:
    return pd.read_csv(io.BytesIO(input_bytes))


def _read_json_bytes(input_bytes: bytes) -> Any:
    return json.loads(input_bytes.decode(DEFAULT_ENCODING))


def read_artifacts(
    storage: Storage,
    input_paths: List[str],
    input_metadata_paths: List[str],
) -> Tuple[List[Any], List[ArtifactType], List[SerializationType]]:
    if len(input_paths) != len(input_metadata_paths):
        raise Exception(
            "Found inconsistent number of input paths (%d) and input metadata paths (%d)"
            % (
                len(input_paths),
                len(input_metadata_paths),
            )
        )

    inputs: List[Any] = []
    artifact_types: List[ArtifactType] = []
    serialization_types: List[SerializationType] = []

    for (input_path, input_metadata_path) in zip(input_paths, input_metadata_paths):
        # Make sure that the input paths exist.
        try:
            _ = storage.get(input_path)
            _ = storage.get(input_metadata_path)
        except Exception as e:
            # TODO(ENG-1627): think about retrying the parent operator in such instances.
            raise MissingInputPathsException(
                "Unable to read inputs artifacts. Exception: %s" % str(e)
            )

        artifact_metadata = json.loads(storage.get(input_metadata_path).decode(DEFAULT_ENCODING))
        artifact_type = artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY]
        artifact_types.append(artifact_type)

        serialization_type = artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY]
        serialization_types.append(serialization_type)

        inputs.append(deserialize(serialization_type, artifact_type, storage.get(input_path)))

    return inputs, artifact_types, serialization_types


def read_system_metadata(
    storage: Storage,
    input_metadata_paths: List[str],
) -> List[Dict[str, Any]]:
    return _read_metadata_key(storage, input_metadata_paths, _METADATA_SYSTEM_METADATA_KEY)


def _read_metadata_key(
    storage: Storage, input_metadata_paths: List[str], key_name: str
) -> List[Dict[str, Any]]:
    metadata_inputs = [
        _read_json_bytes(storage.get(input_path)) for input_path in input_metadata_paths
    ]
    if any(key_name not in metadata for metadata in metadata_inputs):
        raise Exception(key_name + " does not exist in input metadata.")
    return [metadata[key_name] for metadata in metadata_inputs]


def serialize_val_wrapper(val: Any, serialization_type: SerializationType) -> bytes:
    """Wrapper around `serialize_val()` to perform additional checks that are specific to the executor."""
    if serialization_type == SerializationType.TABLE:
        # We cannot serialize integer column names into json.
        violating_col_names = [col for col in val.columns if not isinstance(col, str)]
        if len(violating_col_names) > 0:
            raise ExecFailureException(
                failure_type=FailureType.USER_FATAL,
                tip="Non-String column names are not supported. Violating columns: %s"
                % (", ".join(violating_col_names)),
            )

    serialized_val = serialize_val(val, serialization_type)
    assert isinstance(serialized_val, bytes)  # Necessary for mypy
    return serialized_val


def write_artifact(
    storage: Storage,
    artifact_type: ArtifactType,
    # derived_from_bson specifies if the artifact is derived from a bson object
    # and thus requires bson encoding.
    # For now, it only applies to data frames extracted / transformed from Mongo.
    derived_from_bson: bool,
    output_path: Optional[str],
    output_metadata_path: str,
    content: Any,
    system_metadata: Dict[str, str],
) -> None:
    """The `output_path` can be empty if the contents were already pre-populated (eg. parameter operators)."""
    output_metadata: Dict[str, Any] = {
        _METADATA_SCHEMA_KEY: [],
        _METADATA_SYSTEM_METADATA_KEY: system_metadata,
        _METADATA_ARTIFACT_TYPE_KEY: artifact_type.value,
    }

    if artifact_type == ArtifactType.TABLE:
        output_metadata[_METADATA_SCHEMA_KEY] = [{col: str(content[col].dtype)} for col in content]

    serialization_type = artifact_type_to_serialization_type(
        artifact_type, derived_from_bson, content
    ).value

    if output_path is not None:
        serialized_val = serialize_val_wrapper(content, serialization_type)
        storage.put(output_path, serialized_val)

    output_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = serialization_type
    output_metadata[_METADATA_PYTHON_TYPE_KEY] = type(content).__name__
    storage.put(output_metadata_path, json.dumps(output_metadata).encode(DEFAULT_ENCODING))


def write_exec_state(
    storage: Storage,
    metadata_path: str,
    exec_state: ExecutionState,
) -> None:
    """
    Writes operator execution logs to storage.
    :param err: Any error message encountered during execution.
    :param logs: Any logs generated by this operator.
    """
    storage.put(metadata_path, bytes(exec_state.json(), encoding=DEFAULT_ENCODING))


def delete_object(name: str, delete_fn: Callable[[str], None]) -> SavedObjectDelete:
    exec_state = ExecutionState(user_logs=Logs())
    try:
        delete_fn(name)
    except Exception as e:
        exec_state.status = ExecutionStatus.FAILED
        exec_state.failure_type = FailureType.SYSTEM
        exec_state.error = Error(context=exception_traceback(e), tip=TIP_UNKNOWN_ERROR)
        return SavedObjectDelete(name=name, exec_state=exec_state)
    exec_state.status = ExecutionStatus.SUCCEEDED
    return SavedObjectDelete(name=name, exec_state=exec_state)


def write_delete_saved_objects_results(
    storage: Storage, path: str, results: Dict[str, List[SavedObjectDelete]]
) -> None:
    # `[object].json()`` gives me the json as a string which gets escaped in `json.dumps()``.
    # However, I cannot serialize it directly with `json.dumps()`` so I serialize with
    # `[object].json()` then `json.loads()` to convert it to a dictionary that is serializable
    # by `json.dumps.()`.
    results_str = json.dumps(
        {
            integration: [
                {"name": result.name, "exec_state": json.loads(result.exec_state.json())}
                for result in results[integration]
            ]
            for integration in results
        }
    )
    storage.put(path, bytes(results_str, encoding=DEFAULT_ENCODING))


def write_discover_results(storage: Storage, path: str, tables: List[str]) -> None:
    table_names_str = json.dumps(tables)

    storage.put(path, bytes(table_names_str, encoding=DEFAULT_ENCODING))


def write_compile_airflow_output(storage: Storage, path: str, dag_file: bytes) -> None:
    """
    Writes the provided Airflow DAG file to storage.
    """
    storage.put(path, dag_file)



from aqueduct_executor.operators.connectors.data import config, relational, utils
from sqlalchemy import create_engine, engine


class PostgresConnector(relational.RelationalConnector):
    def __init__(self, config: config.PostgresConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)


def _create_engine(config: config.PostgresConfig) -> engine.Engine:
    # Postgres Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/postgresql.html#module-sqlalchemy.dialects.postgresql.psycopg2
    if config.password:
        url = "postgresql://{username}:{password}@{host}:{port}/{database}".format(
            username=config.username,
            password=utils.url_encode(config.password),
            host=config.host,
            port=config.port,
            database=config.database,
        )
    else:
        url = "postgresql://{username}@{host}:{port}/{database}".format(
            username=config.username,
            host=config.host,
            port=config.port,
            database=config.database,
        )
    return create_engine(url)



from enum import Enum
from typing import Optional, Union

from aqueduct_executor.operators.connectors.data import models
from aqueduct_executor.operators.utils.enums import MetaEnum
from pydantic import Field


class BigQueryConfig(models.BaseConfig):
    project_id: str
    service_account_credentials: str


class MySqlConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: str


class MongoDBConfig(models.BaseConfig):
    auth_uri: str
    database: str


class PostgresConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: Optional[str] = "5432"


class AWSCredentialType(str, Enum, metaclass=MetaEnum):
    ACCESS_KEY = "access_key"
    CONFIG_FILE_PATH = "config_file_path"
    CONFIG_FILE_CONTENT = "config_file_content"


class S3Config(models.BaseConfig):
    # default type to ACCESS_KEY mainly for backward compatibility
    type: AWSCredentialType = AWSCredentialType.ACCESS_KEY

    # Access key credentials
    access_key_id: str = ""
    secret_access_key: str = ""

    # Config credentials
    config_file_path: str = ""
    config_file_content: str = ""
    config_file_profile: str = ""

    bucket: str = ""

    region: str = ""
    use_as_storage: str = ""


class AthenaConfig(models.BaseConfig):
    # default type to ACCESS_KEY mainly for backward compatibility
    type: AWSCredentialType = AWSCredentialType.ACCESS_KEY

    # Access key credentials
    access_key_id: str = ""
    secret_access_key: str = ""
    region: str = ""

    # Config credentials
    config_file_path: str = ""
    config_file_content: str = ""
    config_file_profile: str = ""

    database: str = ""
    output_location: str = ""


class GCSConfig(models.BaseConfig):
    bucket: str
    service_account_credentials: str
    use_as_storage: str = ""


class SnowflakeConfig(models.BaseConfig):
    username: str
    password: str
    account_identifier: str
    database: str
    warehouse: str
    db_schema: Optional[str] = Field("public", alias="schema")  # schema is a Pydantic keyword

    class Config:
        # Ensures that Pydantic parses JSON keys named "schema" or "db_schema" to
        # the `db_schema` field
        allow_population_by_field_name = True


class SqlServerConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: str


class SqliteConfig(models.BaseConfig):
    database: str


Config = Union[
    BigQueryConfig,
    GCSConfig,
    MySqlConfig,
    MongoDBConfig,
    PostgresConfig,
    S3Config,
    AthenaConfig,
    SnowflakeConfig,
    SqlServerConfig,
    SqliteConfig,
]



from __future__ import annotations

import io
import sys
import traceback
from contextlib import redirect_stderr, redirect_stdout
from typing import Any, Callable, Optional, Type, TypeVar

from aqueduct_executor.operators.utils.enums import ExecutionStatus, FailureType
from pydantic import BaseModel

_GITHUB_ISSUE_LINK = "https://github.com/aqueducthq/aqueduct/issues/new?assignees=&labels=bug&template=bug_report.md&title=%5BBUG%5D"

TIP_OP_EXECUTION = "Error executing operator. Please refer to the stack trace for fix."
_TIP_CREATE_BUG_REPORT = (
    f"Please create bug report in github: {_GITHUB_ISSUE_LINK} . "
    "We will get back to you as soon as we can."
)
TIP_UNKNOWN_ERROR = f"Sorry, we've run into an unexpected error! {_TIP_CREATE_BUG_REPORT}"
TIP_INTEGRATION_CONNECTION = (
    "We were unable to connect to this integration. "
    "Please check your credentials or contact your integration's provider."
)
TIP_DEMO_CONNECTION = f"We have trouble connecting to demo DB. {_TIP_CREATE_BUG_REPORT}"

TIP_EXTRACT = "We couldn't execute the provided query. Please double check your query is correct."
TIP_LOAD = "We couldn't load to the integration. Please make sure the target exists, or you have the right permission."
TIP_DISCOVER = "We couldn't list items in the integration. Please make sure your credentials have the right permission."

# Assumption: only check operators will use this tip.
TIP_CHECK_DID_NOT_PASS = "The check did not pass (returned False)."

TIP_NOT_NUMERIC = "The computed result is not of type numeric."
TIP_NOT_BOOL = "The computed result is not of type bool."


class Error(BaseModel):
    tip: str = ""  # Information about how the user could fix the error.
    context: str = ""  # More details about the error. Typically a stack trace.


class Logs(BaseModel):
    stdout: str = ""
    stderr: str = ""


class ExecFailureException(Exception):
    """Can be thrown whenever you want to fail the operator.

    It captures all the information needed to mark an ExecutionState as failed, except for user logs.
    Can be translated into an ExecutionState with `ExecutionState.from_exception()`.
    """

    def __init__(
        self,
        failure_type: FailureType,
        tip: str,
        context: str = "",
    ):
        self.failure_type = failure_type
        self.tip = tip
        self.context = context


class ExecutionState(BaseModel):
    """
    The state to track operator execution. In the future, we may extend this
    to track arbitrary execution.

    `status`: the status of execution, one of 'pending', 'succeeded', or 'failed'.
    `user_logs`: the stderr and stdout of 'user' part of execution. Available regardless of status.
    `failure_type`: more detailed failure reason. Available only if status is `failed`.
    `error`:  structured error message. Available only if status is `failed`.
    """

    user_logs: Logs
    status: ExecutionStatus = ExecutionStatus.PENDING
    failure_type: Optional[FailureType] = None
    error: Optional[Error] = None

    @classmethod
    def from_exception(cls, e: ExecFailureException, user_logs: Logs) -> ExecutionState:
        """Translates a ExecFailureException into a failed ExecutionState we can persist.

        We explicitly require `user_logs`, because the exception does not contain that context, so
        this will force us to explicitly think about when we want to persist logs.
        """
        return cls(
            user_logs=user_logs,
            status=ExecutionStatus.FAILED,
            failure_type=e.failure_type,
            error=Error(
                tip=e.tip,
                context=e.context,
            ),
        )

    def mark_as_failure(self, failure_type: FailureType, tip: str, context: str = "") -> None:
        self.status = ExecutionStatus.FAILED
        self.failure_type = failure_type
        self.error = Error(
            tip=tip,
            context=context,
        )

    def user_fn_redirected(self, failure_tip: str) -> Callable[..., Any]:
        """
        Usage:
        ```
        @exec_state.user_fn_redirected(failure_tip="some message when decorated fn failed")
        def user_fn():
            # run some fn user specified

        user_fn()
        ```
        When decorated with `user_fn_redirected`, the stdout and stderr will be redirected
        to `user_logs`.

        When the decorated fn failed, the `exec_state` will be 'failed' with type 'user'.
        The `error` object will contain the first frame of stack together with the tip provided.
        """

        def wrapper(user_fn: Callable[..., Any]) -> Callable[..., Any]:
            def inner(*args: Any, **kwargs: Any) -> Any:
                stdout_log = io.StringIO()
                stderr_log = io.StringIO()
                try:
                    with redirect_stdout(stdout_log), redirect_stderr(stderr_log):
                        result = user_fn(*args, **kwargs)
                except Exception:
                    # Include the stack trace within the user's code.
                    _set_redirected_logs(stdout_log, stderr_log, self.user_logs)
                    self.mark_as_failure(
                        FailureType.USER_FATAL,
                        failure_tip,
                        context=stack_traceback(
                            offset=1
                        ),  # traceback the first stack frame, which belongs to user
                    )
                    print(f"User failure. Full log: {self.json()}")
                    return None

                # Include the stack trace within the user's code.
                _set_redirected_logs(stdout_log, stderr_log, self.user_logs)
                print(f"User execution succeeded. Full log: {self.json()}")
                return result

            return inner

        return wrapper


def _set_redirected_logs(
    stdout: io.StringIO,
    stderr: io.StringIO,
    logs: Logs,
) -> None:
    """
    If there is any output, set as the values for protected keys STDOUT_KEY and STDERR_KEY.
    """
    stdout.seek(0)
    stderr.seek(0)

    stdout_contents = stdout.read()
    if len(stdout_contents) > 0:
        print(f"StdOut: \n {stdout_contents}")
        logs.stdout = stdout_contents

    stderr_contents = stderr.read()
    if len(stderr_contents) > 0:
        print(f"StdErr: \n {stderr_contents}")
        logs.stderr = stderr_contents
    return


def stack_traceback(offset: int = 0) -> str:
    """
    Captures the stack traceback and returns it as a string. If offset is positive,
    it will extract the traceback starting at OFFSET frames from the top (e.g. most recent frame).
    An offset of 1 means the most recent frame will be excluded.

    This is typically used for user function traceback so that we throw away
    unnecessary stack frames.
    """
    file = io.StringIO()

    tb_type, tb_val, tb = sys.exc_info()
    while offset > 0:
        if tb is None or tb.tb_next is None:
            break
        tb = tb.tb_next
        offset -= 1

    traceback.print_exception(tb_type, tb_val, tb, file=file)

    file.seek(0)
    return file.read()


def exception_traceback(exception: Exception) -> str:
    """
    `exception_traceback` prints the traceback of the entire exception.

    This is typically used for system error so that the full trace is captured.
    """
    return (
        "".join(traceback.format_tb(exception.__traceback__))
        + f"{exception.__class__.__name__}: {str(exception)}"
    )



from aqueduct_executor.operators.utils.storage.config import StorageConfig
from aqueduct_executor.operators.utils.storage.file import FileStorage
from aqueduct_executor.operators.utils.storage.gcs import GCSStorage
from aqueduct_executor.operators.utils.storage.s3 import S3Storage
from aqueduct_executor.operators.utils.storage.storage import Storage


def parse_storage(storage_config: StorageConfig) -> Storage:
    if storage_config.s3_config:
        return S3Storage(storage_config.s3_config)
    if storage_config.file_config:
        return FileStorage(storage_config.file_config)
    if storage_config.gcs_config:
        return GCSStorage(storage_config.gcs_config)
    raise Exception("Unknown storage type")



from pydantic import BaseModel, Extra


class BaseSpec(BaseModel):
    """
    BaseSpec defines the Pydantic Config shared by all connector Spec's, e.g.
    AuthenticateSpec, ExtractSpec, etc.
    """

    class Config:
        extra = Extra.forbid  # Ensures extra fields are not allowed
        smart_union = True  # Prevents undesired coercion does not occur.


class BaseConfig(BaseModel):
    """
    BaseConfig defines the Pydantic Config shared by all connector Config's, e.g.
    postgres.Config, mysql.Config, etc.
    """

    class Config:
        extra = Extra.forbid


class BaseParams(BaseModel):
    """
    BaseParams defines the Pydantic Config shared by all ExtractParams and LoadParams, e.g.
    relational.ExtractParams, relational.LoadParams, etc.
    """

    class Config:
        extra = Extra.forbid



import os
from typing import Any, Tuple

import boto3
from aqueduct_executor.operators.utils.storage.config import S3StorageConfig
from aqueduct_executor.operators.utils.storage.storage import Storage
from botocore.config import Config as BotoConfig


class S3Storage(Storage):
    _client: Any  # boto3 s3 client
    _config: S3StorageConfig

    def __init__(self, config: S3StorageConfig):

        if config.aws_access_key_id and config.aws_secret_access_key:
            # The AWS keys are passed in as part of the storage spec for AWS Lambda engines
            self._client = boto3.client(
                "s3",
                aws_access_key_id=config.aws_access_key_id,
                aws_secret_access_key=config.aws_secret_access_key,
            )
        elif "AWS_ACCESS_KEY_ID" in os.environ and "AWS_SECRET_ACCESS_KEY" in os.environ:
            # The AWS keys are passed in as environment variables for k8s engines
            self._client = boto3.client(
                "s3",
                aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
                aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
            )
        else:
            # Boto3 uses an environment variable to determine the credentials filepath and profile
            os.environ["AWS_SHARED_CREDENTIALS_FILE"] = config.credentials_path
            os.environ["AWS_PROFILE"] = config.credentials_profile
            self._client = boto3.client("s3", config=BotoConfig(region_name=config.region))
        self._config = config

        bucket, key_prefix = parse_s3_path(self._config.bucket)
        self._bucket = bucket
        self._key_prefix = key_prefix

    def put(self, key: str, value: bytes) -> None:
        key = self._prefix_key(key)
        print(f"writing to s3: {key}")
        self._client.put_object(Bucket=self._bucket, Key=key, Body=value)

    def get(self, key: str) -> bytes:
        key = self._prefix_key(key)
        print(f"reading from s3: {key}")
        return self._client.get_object(Bucket=self._bucket, Key=key)["Body"].read()  # type: ignore

    def _prefix_key(self, key: str) -> str:
        if not self._key_prefix:
            return key
        return self._key_prefix + "/" + key


def parse_s3_path(s3_path: str) -> Tuple[str, str]:
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)
    return bucket, key



import pytest
from aqueduct_executor.operators.connectors.data import dataframe, redshift
from aqueduct_executor.operators.connectors.tests import conf, utils

_TABLE = "test_redshift"


@pytest.mark.skipif(conf.SKIP_REDSHIFT, reason="Skip Redshift Flag Set")
class TestPostgres:
    @classmethod
    def setup_class(cls):
        # Setup connector
        config = conf.REDSHIFT_CONF
        conn = redshift.RedshiftConnector(config)
        cls.conn = conn

        # Setup test dataframe
        cls.test_df = utils.sample_df()

    @classmethod
    def teardown_class(cls):
        cls.conn.engine.connect().execute("DROP TABLE IF EXISTS {} CASCADE;".format(_TABLE))

    def test_authenticate(self):
        utils.authenticate_test(self.conn)

    @pytest.mark.dependency()
    def test_load(self):
        params = {dataframe.LOAD_PARAMS_TABLE_KEY: _TABLE}
        utils.load_test(self.conn, params, self.test_df)

    @pytest.mark.dependency(depends=["TestRedshift::test_load"])
    def test_extract(self):
        params = {dataframe.EXTRACT_PARAMS_QUERY_KEY: "SELECT * FROM {};".format(_TABLE)}
        utils.extract_test(self.conn, params, expected_df=self.test_df)



from enum import Enum, EnumMeta
from typing import Any, Iterable, cast


class MetaEnum(EnumMeta):
    """Allows to very easily check if strings are present in the enum, without a helper.
    Eg.
        if "Postgres" in ServiceType:
            ...
    """

    def __contains__(cls, item: Any) -> Any:
        return item in [v.value for v in cast(Iterable[Enum], cls.__members__.values())]


class OperatorType(str, Enum, metaclass=MetaEnum):
    FUNCTION = "function"
    METRIC = "metric"
    CHECK = "check"
    EXTRACT = "extract"
    LOAD = "load"
    PARAM = "param"
    SYSTEM_METRIC = "system_metric"


class CheckSeverity(str, Enum, metaclass=MetaEnum):
    ERROR = "error"
    WARNING = "warning"


class ExecutionStatus(str, Enum, metaclass=MetaEnum):
    UNKNOWN = "unknown"
    PENDING = "pending"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class FailureType(Enum, metaclass=MetaEnum):
    SYSTEM = 1
    USER_FATAL = 2
    # For failures that don't stop execution.
    # Eg. check operator with WARNING severity fails.
    USER_NON_FATAL = 3


class JobType(str, Enum, metaclass=MetaEnum):
    FUNCTION = "function"
    AUTHENTICATE = "authenticate"
    EXTRACT = "extract"
    LOAD = "load"
    LOADTABLE = "load-table"
    DELETESAVEDOBJECTS = "delete-saved-objects"
    DISCOVER = "discover"
    PARAM = "param"
    SYSTEM_METRIC = "system_metric"
    COMPILE_AIRFLOW = "compile_airflow"


class ArtifactType(str, Enum, metaclass=MetaEnum):
    UNTYPED = "untyped"
    STRING = "string"
    BOOL = "boolean"
    NUMERIC = "numeric"
    DICT = "dictionary"
    TUPLE = "tuple"
    LIST = "list"
    TABLE = "table"
    JSON = "json"
    BYTES = "bytes"
    IMAGE = "image"
    PICKLABLE = "picklable"
    TF_KERAS = "tensorflow-keras-model"


class SerializationType(str, Enum, metaclass=MetaEnum):
    TABLE = "table"
    BSON_TABLE = "bson_table"
    JSON = "json"
    PICKLE = "pickle"
    IMAGE = "image"
    STRING = "string"
    BYTES = "bytes"
    TF_KERAS = "tensorflow-keras-model"



import json

from aqueduct_executor.operators.utils.storage import config
from pydantic import BaseModel, parse_obj_as


class MigrationSpec(BaseModel):
    param_val: str
    param_type: str
    op: str


def parse_spec(spec_json: bytes) -> MigrationSpec:
    """
    Parses a JSON string into a MigrationSpec.
    """
    data = json.loads(spec_json)
    return parse_obj_as(MigrationSpec, data)



import os
import sys

PYTHON_VERSION_FILE = "python_version.txt"

# Conda environments
PY_37 = "py37"
PY_38 = "py38"
PY_39 = "py39"
PY_310 = "py310"


if __name__ == "__main__":
    # Change to the path where PYTHON_VERSION_FILE resides.
    # The path is input from the calling bash script.
    os.chdir(sys.argv[1])

    if os.path.isfile(PYTHON_VERSION_FILE):
        with open(PYTHON_VERSION_FILE, "r") as f:
            full_python_version = f.read()
            python_version = full_python_version.split(" ")[-1].strip()
            if python_version == "3.10":
                conda_env = PY_310
            elif python_version == "3.9":
                conda_env = PY_39
            elif python_version == "3.8":
                conda_env = PY_38
            elif python_version == "3.7":
                conda_env = PY_37
            else:
                raise Exception("Unsupported Python version: %s" % python_version)
    else:
        # For backwards compatabiliy. Ensures that workflows without a python_version.txt still work.
        conda_env = PY_38

    # The output of the print statement to stdout is captured by the calling bash script into a variable,
    # so we should not include any other print statements in this Python script.
    print(conda_env)



import json
import uuid
from typing import Dict, List, Literal, Union

from aqueduct_executor.operators.connectors.data import spec as conn_spec
from aqueduct_executor.operators.function_executor import spec as func_spec
from aqueduct_executor.operators.param_executor import spec as param_spec
from aqueduct_executor.operators.utils import enums
from aqueduct_executor.operators.utils.storage import config
from pydantic import BaseModel, parse_obj_as

OperatorSpec = Union[
    conn_spec.ExtractSpec, conn_spec.LoadSpec, func_spec.FunctionSpec, param_spec.ParamSpec
]


class CompileAirflowSpec(BaseModel):
    name: str
    type: Literal[enums.JobType.COMPILE_AIRFLOW]
    storage_config: config.StorageConfig
    metadata_path: str
    workflow_dag_id: uuid.UUID
    output_content_path: str
    dag_id: str
    cron_schedule: str
    task_specs: Dict[str, OperatorSpec]
    task_edges: Dict[str, List[str]]


def parse_spec(spec_json: bytes) -> CompileAirflowSpec:
    """
    Parses a JSON string into a CompileAirflowSpec.
    """
    data = json.loads(spec_json)

    return parse_obj_as(CompileAirflowSpec, data)



from enum import Enum

from aqueduct_executor.operators.utils import enums


class Name(Enum, metaclass=enums.MetaEnum):
    POSTGRES = "Postgres"
    SNOWFLAKE = "Snowflake"
    BIG_QUERY = "BigQuery"
    REDSHIFT = "Redshift"
    SQL_SERVER = "SQL Server"
    MYSQL = "MySQL"
    MARIA_DB = "MariaDB"
    AZURE_SQL = "AzureSQL"
    S3 = "S3"
    ATHENA = "Athena"
    SQLITE = "SQLite"
    AQUEDUCT_DEMO = "Aqueduct Demo"
    GCS = "GCS"
    MONGO_DB = "MongoDB"


class UpdateMode(Enum, metaclass=enums.MetaEnum):
    APPEND = "append"
    REPLACE = "replace"
    FAIL = "fail"


class S3TableFormat(Enum, metaclass=enums.MetaEnum):
    JSON = "JSON"
    CSV = "CSV"
    PARQUET = "Parquet"



import base64
import json

from aqueduct_executor.migrators.backfill_python_type_000022 import serialize
from aqueduct_executor.migrators.backfill_python_type_000022.spec import MigrationSpec
from aqueduct_executor.operators.utils.storage.parse import parse_storage


def run(spec: MigrationSpec) -> None:
    storage = parse_storage(spec.storage_config)
    content = storage.get(spec.content_path)

    deserialized_content = serialize.deserialization_function_mapping[spec.serialization_type](
        content
    )
    print(type(deserialized_content).__name__)



from abc import ABC, abstractmethod


class Storage(ABC):
    @abstractmethod
    def put(self, key: str, value: bytes) -> None:
        pass

    @abstractmethod
    def get(self, key: str) -> bytes:
        pass



from enum import Enum
from typing import Optional

from aqueduct_executor.operators.utils.enums import MetaEnum
from pydantic import BaseModel


class StorageType(str, Enum, metaclass=MetaEnum):
    S3 = "s3"
    File = "file"
    GCS = "gcs"


class FileStorageConfig(BaseModel):
    directory: str


class S3StorageConfig(BaseModel):
    region: str
    bucket: str
    credentials_path: str
    credentials_profile: str
    aws_access_key_id: str = ""
    aws_secret_access_key: str = ""


class GCSStorageConfig(BaseModel):
    bucket: str
    service_account_credentials: str


class StorageConfig(BaseModel):
    type: StorageType
    file_config: Optional[FileStorageConfig] = None
    s3_config: Optional[S3StorageConfig] = None
    gcs_config: Optional[GCSStorageConfig] = None



import os
import urllib.parse
import uuid
from typing import Union

import boto3
from aqueduct_executor.operators.connectors.data.config import (
    AthenaConfig,
    AWSCredentialType,
    S3Config,
)


def _session_from_config_file_path(config: Union[S3Config, AthenaConfig]) -> boto3.session.Session:
    """
    returns a boto session that sets access_key and secret_access_key
    based on credentials in config_file_path and config_file_profile.
    """
    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = config.config_file_path
    os.environ["AWS_CONFIG_FILE"] = config.config_file_path
    session = boto3.Session(profile_name=config.config_file_profile)
    # This ensures the credential is cached by the session in memory, so even if the temp credential
    # file is removed, we can still access AWS resources.
    session.get_credentials()
    return session


def _session_from_config_file_content(
    config: Union[S3Config, AthenaConfig]
) -> boto3.session.Session:
    """
    returns a boto session that sets access_key and secret_access_key
    based on credentials in config_file_content and config_file_profile.
    """
    # write to temp file assuming the cwd is safe to create such file.
    temp_path = os.path.join(os.getcwd(), str(uuid.uuid4()))
    with open(temp_path, "w") as w:
        w.write(config.config_file_content)
    config.config_file_path = temp_path

    try:
        return _session_from_config_file_path(config)
    finally:
        # always remove the temp file.
        os.remove(temp_path)


def _session_from_access_key(config: Union[S3Config, AthenaConfig]) -> boto3.session.Session:
    if config.region == "":
        # This is a defensive fallback in case the region field of `S3Config` is empty.
        return boto3.Session(
            aws_access_key_id=config.access_key_id,
            aws_secret_access_key=config.secret_access_key,
        )
    else:
        return boto3.Session(
            aws_access_key_id=config.access_key_id,
            aws_secret_access_key=config.secret_access_key,
            region_name=config.region,
        )


def construct_boto_session(config: Union[S3Config, AthenaConfig]) -> boto3.session.Session:
    if config.type == AWSCredentialType.CONFIG_FILE_CONTENT:
        # Write a temp file
        return _session_from_config_file_content(config)
    elif config.type == AWSCredentialType.CONFIG_FILE_PATH:
        return _session_from_config_file_path(config)
    elif config.type == AWSCredentialType.ACCESS_KEY:
        return _session_from_access_key(config)
    else:
        raise Exception("Unsupported integration config type: %s" % config.type)


def url_encode(value: str) -> str:
    return urllib.parse.quote_plus(value)



import sys
from typing import Any

from aqueduct_executor.operators.connectors.data import common, config, connector, extract
from aqueduct_executor.operators.connectors.data.spec import (
    AQUEDUCT_DEMO_NAME,
    AuthenticateSpec,
    DiscoverSpec,
    ExtractSpec,
    LoadSpec,
    LoadTableSpec,
    Spec,
)
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    ExecutionStatus,
    FailureType,
    JobType,
    SerializationType,
)
from aqueduct_executor.operators.utils.exceptions import MissingConnectorDependencyException
from aqueduct_executor.operators.utils.execution import (
    TIP_DEMO_CONNECTION,
    TIP_EXTRACT,
    TIP_INTEGRATION_CONNECTION,
    TIP_LOAD,
    TIP_UNKNOWN_ERROR,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.storage.storage import Storage


def run(spec: Spec) -> None:
    """
    Runs one of the following connector operations:
    - authenticate
    - extract
    - load
    - load-table
    - delete-saved-objects
    - discover

    Arguments:
    - spec: The spec provided for this operator.
    """
    print("Started %s job: %s" % (spec.type, spec.name))

    storage = parse_storage(spec.storage_config)
    exec_state = ExecutionState(user_logs=Logs())

    try:
        _execute(spec, storage, exec_state)
        # Write operator execution metadata
        # Each decorator may set exec_state.status to FAILED, but if none of them did, then we are
        # certain that the operator succeeded.
        if exec_state.status == ExecutionStatus.FAILED:
            print(f"Failed with error. Full Logs:\n{exec_state.json()}")
            utils.write_exec_state(storage, spec.metadata_path, exec_state)
            sys.exit(1)

        exec_state.status = ExecutionStatus.SUCCEEDED
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
    except ExecFailureException as e:
        # We must reconcile the user logs here, since those logs are not captured on the exception.
        from_exception_exec_state = ExecutionState.from_exception(e, user_logs=exec_state.user_logs)

        print(f"Failed with error. Full Logs:\n{from_exception_exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, from_exception_exec_state)
        sys.exit(1)
    except MissingConnectorDependencyException as e:
        exec_state.mark_as_failure(
            FailureType.USER_FATAL, tip=str(e), context=exception_traceback(e)
        )
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)
    except Exception as e:
        exec_state.mark_as_failure(
            FailureType.SYSTEM, tip=TIP_UNKNOWN_ERROR, context=exception_traceback(e)
        )
        print(f"Failed with system error. Full Logs:\n{exec_state.json()}")

        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)


def _execute(spec: Spec, storage: Storage, exec_state: ExecutionState) -> None:

    if isinstance(spec.connector_name, dict):
        run_delete_saved_objects(spec, storage, exec_state)
    else:
        # Because constructing certain connectors (eg. Postgres) can also involve authentication,
        # we do both in `run_authenticate()`, and give a more helpful error message on failure.
        if spec.type == JobType.AUTHENTICATE:
            run_authenticate(spec, exec_state, is_demo=(spec.name == AQUEDUCT_DEMO_NAME))
            return

        op = setup_connector(spec.connector_name, spec.connector_config)
        if spec.type == JobType.EXTRACT:
            run_extract(spec, op, storage, exec_state)
        elif spec.type == JobType.LOADTABLE:
            run_load_table(spec, op, storage)
        elif spec.type == JobType.LOAD:
            run_load(spec, op, storage, exec_state)
        elif spec.type == JobType.DISCOVER:
            run_discover(spec, op, storage)
        else:
            raise Exception("Unknown job: %s" % spec.type)


def run_authenticate(
    spec: AuthenticateSpec,
    exec_state: ExecutionState,
    is_demo: bool,
) -> None:
    @exec_state.user_fn_redirected(
        failure_tip=TIP_DEMO_CONNECTION if is_demo else TIP_INTEGRATION_CONNECTION
    )
    def _authenticate() -> None:
        op = setup_connector(spec.connector_name, spec.connector_config)
        op.authenticate()

    _authenticate()


def run_extract(
    spec: ExtractSpec, op: connector.DataConnector, storage: Storage, exec_state: ExecutionState
) -> None:
    extract_params = spec.parameters

    # Search for user-defined placeholder if this is a relational query, and replace them with
    # the appropriate values.
    if isinstance(extract_params, extract.RelationalParams) or isinstance(
        extract_params, extract.MongoDBParams
    ):
        assert len(spec.input_param_names) == len(spec.input_content_paths)
        input_vals, _, _ = utils.read_artifacts(
            storage,
            spec.input_content_paths,
            spec.input_metadata_paths,
        )
        assert all(
            isinstance(param_val, str) for param_val in input_vals
        ), "Parameter value must be a string."

        parameters = dict(zip(spec.input_param_names, input_vals))
        extract_params.compile(parameters)

    @exec_state.user_fn_redirected(failure_tip=TIP_EXTRACT)
    def _extract() -> Any:
        return op.extract(spec.parameters)

    output = _extract()

    output_artifact_type = ArtifactType.TABLE
    derived_from_bson = isinstance(extract_params, extract.MongoDBParams)
    if isinstance(extract_params, extract.S3Params):
        output_artifact_type = extract_params.artifact_type
        # If the type of the output is tuple, then it could be a multi-file S3 request so we
        # overwrite the output type to tuple.
        if isinstance(output, tuple):
            output_artifact_type = ArtifactType.TUPLE

    if exec_state.status != ExecutionStatus.FAILED:
        utils.write_artifact(
            storage,
            output_artifact_type,
            derived_from_bson,
            spec.output_content_path,
            spec.output_metadata_path,
            output,
            system_metadata={},
        )


def run_delete_saved_objects(spec: Spec, storage: Storage, exec_state: ExecutionState) -> None:
    results = {}
    assert isinstance(spec.connector_name, dict)
    for integration in spec.connector_name:
        op = setup_connector(spec.connector_name[integration], spec.connector_config[integration])
        results[integration] = op.delete(spec.integration_to_object[integration])
    utils.write_delete_saved_objects_results(storage, spec.output_content_path, results)


def run_load(
    spec: LoadSpec, op: connector.DataConnector, storage: Storage, exec_state: ExecutionState
) -> None:
    inputs, input_types, _ = utils.read_artifacts(
        storage,
        [spec.input_content_path],
        [spec.input_metadata_path],
    )
    if len(inputs) != 1:
        raise Exception("Expected 1 input artifact, but got %d" % len(inputs))

    @exec_state.user_fn_redirected(failure_tip=TIP_LOAD)
    def _load() -> None:
        op.load(spec.parameters, inputs[0], input_types[0])

    _load()


def run_load_table(spec: LoadTableSpec, op: connector.DataConnector, storage: Storage) -> None:
    df = utils._read_csv(storage.get(spec.csv))
    op.load(spec.load_parameters.parameters, df, ArtifactType.TABLE)


def run_discover(spec: DiscoverSpec, op: connector.DataConnector, storage: Storage) -> None:
    tables = op.discover()
    utils.write_discover_results(storage, spec.output_content_path, tables)


def setup_connector(
    connector_name: common.Name, connector_config: config.Config
) -> connector.DataConnector:
    # prevent isort from moving around type: ignore comments which will cause mypy issues.
    # isort: off
    if connector_name == common.Name.AQUEDUCT_DEMO or connector_name == common.Name.POSTGRES:
        try:
            import psycopg2
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Postgres connector. Have you run `aqueduct install postgres`?"
            )

        from aqueduct_executor.operators.connectors.data.postgres import (
            PostgresConnector as OpConnector,
        )
    elif connector_name == common.Name.SNOWFLAKE:
        try:
            from snowflake import sqlalchemy
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Snowflake connector. Have you run `aqueduct install snowflake`?"
            )

        from aqueduct_executor.operators.connectors.data.snowflake import (  # type: ignore
            SnowflakeConnector as OpConnector,
        )
    elif connector_name == common.Name.BIG_QUERY:
        try:
            from google.cloud import bigquery
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the BigQuery connector. Have you run `aqueduct install bigquery`?"
            )

        from aqueduct_executor.operators.connectors.data.bigquery import (  # type: ignore
            BigQueryConnector as OpConnector,
        )
    elif connector_name == common.Name.REDSHIFT:
        try:
            import psycopg2
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Redshift connector. Have you run `aqueduct install redshift`?"
            )

        from aqueduct_executor.operators.connectors.data.redshift import (  # type: ignore
            RedshiftConnector as OpConnector,
        )
    elif connector_name == common.Name.SQL_SERVER:
        try:
            import pyodbc
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the SQL Server connector. Have you run `aqueduct install sqlserver`?"
            )

        from aqueduct_executor.operators.connectors.data.sql_server import (  # type: ignore
            SqlServerConnector as OpConnector,
        )
    elif connector_name == common.Name.MYSQL:
        try:
            import MySQLdb
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the MySQL connector. Have you run `aqueduct install mysql`?"
            )

        from aqueduct_executor.operators.connectors.data.mysql import (  # type: ignore
            MySqlConnector as OpConnector,
        )
    elif connector_name == common.Name.MARIA_DB:
        try:
            import MySQLdb
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the MariaDB connector. Have you run `aqueduct install mariadb`?"
            )

        from aqueduct_executor.operators.connectors.data.maria_db import (  # type: ignore
            MariaDbConnector as OpConnector,
        )
    elif connector_name == common.Name.AZURE_SQL:
        try:
            import pyodbc
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Azure SQL connector. Have you run `aqueduct install azuresql`?"
            )

        from aqueduct_executor.operators.connectors.data.azure_sql import (  # type: ignore
            AzureSqlConnector as OpConnector,
        )
    elif connector_name == common.Name.S3:
        try:
            import pyarrow
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the S3 connector. Have you run `aqueduct install s3`?"
            )

        from aqueduct_executor.operators.connectors.data.s3 import (  # type: ignore
            S3Connector as OpConnector,
        )
    elif connector_name == common.Name.SQLITE:
        from aqueduct_executor.operators.connectors.data.sqlite import (  # type: ignore
            SqliteConnector as OpConnector,
        )
    elif connector_name == common.Name.GCS:
        from aqueduct_executor.operators.connectors.data.gcs import (  # type: ignore
            GCSConnector as OpConnector,
        )
    elif connector_name == common.Name.ATHENA:
        try:
            import awswrangler
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize the Athena connector. Have you run `aqueduct install athena`?"
            )

        from aqueduct_executor.operators.connectors.data.athena import (  # type: ignore
            AthenaConnector as OpConnector,
        )
    elif connector_name == common.Name.MONGO_DB:
        try:
            import pymongo
        except:
            raise MissingConnectorDependencyException(
                "Unable to initialize MongoDB connector. Have you run `aqueduct install mongodb`?"
            )

        from aqueduct_executor.operators.connectors.data.mongodb import (  # type: ignore
            MongoDBConnector as OpConnector,
        )
    else:
        raise Exception("Unknown connector name: %s" % connector_name)

    # isort: on
    return OpConnector(config=connector_config)  # type: ignore



import json
from typing import Any, List

from aqueduct_executor.operators.connectors.data import connector
from aqueduct_executor.operators.connectors.data.config import GCSConfig
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from google.cloud import storage
from google.oauth2 import service_account

_CREDENTIALS_ENV_VAR = "GCS_CREDENTIALS"


class GCSConnector(connector.DataConnector):
    _client: Any  # GCS client
    _config: GCSConfig
    _temp_credentials_path: str = ""

    def __init__(self, config: GCSConfig):
        credentials_info = json.loads(config.service_account_credentials)
        credentials = service_account.Credentials.from_service_account_info(credentials_info)
        self._client = storage.Client(credentials=credentials)
        self._config = config

    def authenticate(self) -> None:
        self._client.list_buckets()

    def discover(self) -> List[str]:
        raise Exception("Discover is not supported for GCS.")

    def extract(self, params: Any) -> Any:
        raise Exception("Extract is not currently supported for GCS.")

    def load(self, params: Any, data: Any, artifact_type: ArtifactType) -> None:
        raise Exception("Load is not currently supported for GCS.")

    def delete(self, objects: Any) -> List[SavedObjectDelete]:
        raise Exception("Delete is not currently supported for GCS.")



import argparse
import base64
import subprocess
import sys
from typing import Optional

from aqueduct_executor.operators.function_executor.spec import FunctionSpec, parse_spec
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import FailureType
from aqueduct_executor.operators.utils.execution import ExecFailureException, ExecutionState, Logs
from aqueduct_executor.operators.utils.storage.parse import parse_storage


def install_missing_packages(
    missing_path: str, spec: FunctionSpec, conda_env: Optional[str]
) -> None:

    if conda_env:
        install_output = subprocess.run(
            [
                "conda",
                "run",
                "-n",
                conda_env,
                sys.executable,
                "-m",
                "pip",
                "install",
                "-r",
                missing_path,
            ],
            capture_output=True,
            text=True,
        )
    else:
        install_output = subprocess.run(
            [sys.executable, "-m", "pip", "install", "-r", missing_path],
            capture_output=True,
            text=True,
        )

    if install_output.returncode != 0:
        exception = ExecFailureException(
            failure_type=FailureType.USER_FATAL,
            tip="We are unable to install certain dependency packages. Please remove them from the \
requirement file and try again. Please refer to the stderr log for which package \
caused the installation error.",
        )
        from_exception_exec_state = ExecutionState.from_exception(
            exception, user_logs=Logs(stdout=install_output.stdout, stderr=install_output.stderr)
        )

        utils.write_exec_state(
            parse_storage(spec.storage_config), spec.metadata_path, from_exception_exec_state
        )
        sys.exit(1)


def run(
    local_path: str,
    requirements_path: str,
    missing_path: str,
    spec: FunctionSpec,
    conda_env: Optional[str] = None,
) -> None:
    with open(local_path, "r") as f:
        local_req = set(f.read().split("\n"))

    with open(requirements_path, "r") as f:
        required = f.read().split("\n")

    missing = []
    for r in required:
        # Remove any @ file because we may not have those files local to the user's device in our file system.
        if r not in local_req and "@ file" not in r:
            missing.append(r)

    if len(missing) > 0:
        with open(missing_path, "w") as f:
            f.write("\n".join(missing))
        install_missing_packages(missing_path, spec, conda_env)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--local_path", required=True)
    parser.add_argument("--requirements_path", required=True)
    parser.add_argument("--missing_path", required=True)
    parser.add_argument("--spec", required=True)
    parser.add_argument("--conda_env")
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = parse_spec(spec_json)

    run(args.local_path, args.requirements_path, args.missing_path, spec, args.conda_env)



import pandas as pd
import pytest
from aqueduct_executor.operators.connectors.data import connector


def authenticate_test(conn: connector.DataConnector):
    try:
        conn.authenticate()
    except ConnectionError as e:
        pytest.fail("Failed authentication %s" % e)


def load_test(conn: connector.DataConnector, params: dict, df: pd.DataFrame):
    conn.load(params, df)


def extract_test(conn: connector.DataConnector, params: dict, expected_df: pd.DataFrame):
    df = conn.extract(params)
    dup = pd.concat([df, expected_df]).drop_duplicates(keep=False)
    if dup.shape[0] != 0:
        pytest.fail(
            "Extracted dataframe does not match expected dataframe.\n Actual DF:\n {}\n Expected DF:\n {}\n".format(
                df, expected_df
            )
        )


def sample_df() -> pd.DataFrame:
    return pd.read_csv("https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv")



from aqueduct_executor.operators.utils.execution import ExecutionState
from pydantic import BaseModel


class SavedObjectDelete(BaseModel):
    """This contains the result of deleting the saved object."""

    name: str
    exec_state: ExecutionState



from typing import Optional, Union

from aqueduct_executor.operators.connectors.data import common, models
from pydantic import validator


class RelationalParams(models.BaseParams):
    table: str
    update_mode: common.UpdateMode = common.UpdateMode.REPLACE

    class Config:
        validate_assignment = True

    @validator("update_mode")
    def set_update_mode(cls, update_mode):  # type: ignore
        if update_mode == "":
            return common.UpdateMode.REPLACE
        return update_mode


class S3Params(models.BaseParams):
    filepath: str
    format: Optional[common.S3TableFormat]


Params = Union[RelationalParams, S3Params]



import os
import sys

_REQUIREMENTS_FILE = "./python/aqueduct_executor/operators/connectors/tests/requirements.txt"


def pytest_configure(config):
    # Install required packages
    os.system(f"{sys.executable} -m pip install -r {_REQUIREMENTS_FILE}")



import json
from typing import Any, Dict, List, Optional

import pandas as pd
from aqueduct_executor.operators.connectors.data import common, config, connector, extract, load
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from google.cloud import bigquery
from google.oauth2 import service_account


class BigQueryConnector(connector.DataConnector):
    def __init__(self, config: config.BigQueryConfig):
        self.project_id = config.project_id

        credentials_info = json.loads(config.service_account_credentials)
        self.credentials = service_account.Credentials.from_service_account_info(credentials_info)
        self.client = bigquery.Client(credentials=self.credentials, project=self.project_id)

    def authenticate(self) -> None:
        # There is no explicit authenticate method, so we execute a test query
        self.client.query("SELECT 1;")

    def discover(self) -> List[str]:
        all_tables = []
        for dataset in self.client.list_datasets():
            tables = self.client.list_tables(dataset.dataset_id)
            all_tables.extend([table.full_table_id.split(":")[-1] for table in tables])
        return all_tables

    def extract(self, params: extract.RelationalParams) -> Any:
        assert params.usable(), "Query is not usable. Did you forget to expand placeholders?"
        query = self.client.query(params.query)
        df = query.result().to_dataframe()
        return df

    def _delete_object(self, name: str, context: Optional[Dict[str, Any]] = None) -> None:
        self.client.delete_table(name, not_found_ok=False)

    def delete(self, tables: List[str]) -> List[SavedObjectDelete]:
        results = []
        for table in tables:
            results.append(delete_object(table, self._delete_object))
        return results

    def load(self, params: load.RelationalParams, df: Any, artifact_type: ArtifactType) -> None:
        if artifact_type != ArtifactType.TABLE:
            raise Exception("The data being loaded must be of type table, found %s" % artifact_type)

        update_mode = params.update_mode
        write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE  # Default
        if update_mode == common.UpdateMode.APPEND:
            write_disposition = bigquery.WriteDisposition.WRITE_APPEND
        if update_mode == common.UpdateMode.REPLACE:
            write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        if update_mode == common.UpdateMode.FAIL:
            write_disposition = bigquery.WriteDisposition.WRITE_EMPTY
        # Since string columns use the "object" dtype, pass in a (partial) schema
        # to ensure the correct BigQuery data type.
        partial_schema = []
        for column in df:
            if df[column].dtype == object:
                partial_schema.append(bigquery.SchemaField(column, "STRING"))

        job_config = bigquery.LoadJobConfig(
            schema=partial_schema, write_disposition=write_disposition
        )
        job = self.client.load_table_from_dataframe(df, params.table, job_config=job_config)

        # Wait for the load job to complete.
        job.result()



import base64
import io
import json
from typing import Any, Callable, Dict, cast

import cloudpickle as pickle
import pandas as pd
from aqueduct_executor.operators.utils.enums import ArtifactType, SerializationType
from PIL import Image

_DEFAULT_ENCODING = "utf8"
_DEFAULT_IMAGE_FORMAT = "jpeg"


def _read_table_content(content: bytes) -> pd.DataFrame:
    return pd.read_json(io.BytesIO(content), orient="table")


def _read_json_content(content: bytes) -> Any:
    return json.loads(content.decode(_DEFAULT_ENCODING))


def _read_pickle_content(content: bytes) -> Any:
    return pickle.loads(content)


def _read_image_content(content: bytes) -> Image.Image:
    return Image.open(io.BytesIO(content))


def _read_string_content(content: bytes) -> str:
    return content.decode(_DEFAULT_ENCODING)


def _read_bytes_content(content: bytes) -> bytes:
    return content


deserialization_function_mapping: Dict[str, Callable[[bytes], Any]] = {
    SerializationType.TABLE: _read_table_content,
    SerializationType.JSON: _read_json_content,
    SerializationType.PICKLE: _read_pickle_content,
    SerializationType.IMAGE: _read_image_content,
    SerializationType.STRING: _read_string_content,
    SerializationType.BYTES: _read_bytes_content,
}


def _write_table_output(output: pd.DataFrame) -> bytes:
    output_str = cast(str, output.to_json(orient="table", date_format="iso", index=False))
    return output_str.encode(_DEFAULT_ENCODING)


def _write_image_output(output: Image.Image) -> bytes:
    img_bytes = io.BytesIO()
    output.save(img_bytes, format=_DEFAULT_IMAGE_FORMAT)
    return img_bytes.getvalue()


def _write_string_output(output: str) -> bytes:
    return output.encode(_DEFAULT_ENCODING)


def _write_bytes_output(output: bytes) -> bytes:
    return output


def _write_pickle_output(output: Any) -> bytes:
    return bytes(pickle.dumps(output))


def _write_json_output(output: Any) -> bytes:
    return json.dumps(output).encode(_DEFAULT_ENCODING)


serialization_function_mapping: Dict[str, Callable[..., bytes]] = {
    SerializationType.TABLE: _write_table_output,
    SerializationType.JSON: _write_json_output,
    SerializationType.PICKLE: _write_pickle_output,
    SerializationType.IMAGE: _write_image_output,
    SerializationType.STRING: _write_string_output,
    SerializationType.BYTES: _write_bytes_output,
}


def serialize_val(val: Any, serialization_type: SerializationType) -> str:
    val_bytes = serialization_function_mapping[serialization_type](val)
    return _bytes_to_base64_string(val_bytes)


def _bytes_to_base64_string(content: bytes) -> str:
    """Helper to convert any bytes-type to a string, by first encoding it with base64 it.
    For example, image-serialized bytes are not `utf8` encoded, so if we want to convert
    such bytes to string, we must use this function.
    """
    return base64.b64encode(content).decode(_DEFAULT_ENCODING)


def artifact_type_to_serialization_type(
    artifact_type: ArtifactType, content: Any
) -> SerializationType:
    """Copy of the same method on in aqueduct executor."""
    if artifact_type == ArtifactType.TABLE:
        serialization_type = SerializationType.TABLE
    elif artifact_type == ArtifactType.IMAGE:
        serialization_type = SerializationType.IMAGE
    elif artifact_type == ArtifactType.JSON or artifact_type == ArtifactType.STRING:
        serialization_type = SerializationType.STRING
    elif artifact_type == ArtifactType.BYTES:
        serialization_type = SerializationType.BYTES
    elif artifact_type == ArtifactType.BOOL or artifact_type == ArtifactType.NUMERIC:
        serialization_type = SerializationType.JSON
    elif artifact_type == ArtifactType.PICKLABLE:
        serialization_type = SerializationType.PICKLE
    elif artifact_type == ArtifactType.DICT or artifact_type == ArtifactType.TUPLE:
        try:
            json.dumps(content)
            serialization_type = SerializationType.JSON
        except:
            serialization_type = SerializationType.PICKLE
    else:
        raise Exception("Unsupported artifact type %s" % artifact_type)

    assert serialization_type is not None
    return serialization_type



import argparse
import base64

from aqueduct_executor.operators.param_executor import execute
from aqueduct_executor.operators.param_executor.spec import parse_spec

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = parse_spec(spec_json)

    execute.run(spec)



import importlib
import json
import os
import shutil
import sys
import tracemalloc
import uuid
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import pandas as pd
from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.operators.function_executor import extract_function, get_extract_path
from aqueduct_executor.operators.function_executor.spec import FunctionSpec
from aqueduct_executor.operators.function_executor.utils import OP_DIR
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import (
    ArtifactType,
    CheckSeverity,
    ExecutionStatus,
    FailureType,
    OperatorType,
    SerializationType,
)
from aqueduct_executor.operators.utils.execution import (
    TIP_CHECK_DID_NOT_PASS,
    TIP_NOT_BOOL,
    TIP_NOT_NUMERIC,
    TIP_OP_EXECUTION,
    TIP_UNKNOWN_ERROR,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.timer import Timer


def _get_py_import_path(spec: FunctionSpec) -> str:
    """
    Generates the import path based on fixed function dir and
    FUNCTION_ENTRY_POINT_FILE env var.

    It removes .py (if any) from the entry point and replaces all
    '/' with '.'

    For example, entry point 'model/churn.py'  will finally become
    'app.function.model.churn', where we can import from.
    """
    file_path = spec.entry_point_file
    if file_path.endswith(".py"):
        file_path = file_path[:-3]

    if file_path.startswith("/"):
        file_path = file_path[1:]
    return file_path.replace("/", ".")


def _import_invoke_method(spec: FunctionSpec) -> Callable[..., Any]:
    """
    `_import_invoke_method` imports the model object.
    it assumes the operator has been extracted to `<storage>/operators/<id>/op`
    and imports the route from the above path.
    """

    # fn_path should be `<storage>/operators/<id>`
    fn_path = spec.function_extract_path

    # work_dir should be `<storage>/operators/<id>/op`
    work_dir = os.path.join(fn_path, OP_DIR)
    print(f"listdir(workdir): {os.listdir(work_dir)}")
    print(f"listdir(fn_path): {os.listdir(fn_path)}")

    # this ensures any file manipulation happens with respect to work_dir
    os.chdir(work_dir)
    # adds work_dir to sys.path to support relative imports from work_dir
    sys.path.append(work_dir)

    import_path = _get_py_import_path(spec)
    print(f"import_path: {import_path}")
    class_name = spec.entry_point_class
    method_name = spec.entry_point_method
    custom_args_str = spec.custom_args

    # Invoke the function and parse out the result object.

    module = importlib.import_module(import_path)

    if not class_name:
        return getattr(module, method_name)  # type: ignore

    fn_class = getattr(module, class_name)
    function = fn_class()
    # Set the custom arguments if provided
    if custom_args_str:
        custom_args = json.loads(custom_args_str)
        function.set_args(custom_args)

    return getattr(function, method_name)  # type: ignore


def _execute_function(
    spec: FunctionSpec,
    inputs: List[Any],
    exec_state: ExecutionState,
) -> Tuple[List[Any], List[ArtifactType], Dict[str, str]]:
    """
    Invokes the given function on the input data. Does not raise an exception on any
    user function errors, but instead annotates the given exec state with the error.

    :param inputs: the input data to feed into the user's function.
    """

    invoke = _import_invoke_method(spec)
    timer = Timer()
    print("Invoking the function...")
    timer.start()
    tracemalloc.start()

    @exec_state.user_fn_redirected(failure_tip=TIP_OP_EXECUTION)
    def _invoke() -> Any:
        return invoke(*inputs)

    results = _invoke()

    # We only validate the number of results if multiple outputs are expected.
    # Otherwise, we treat the results as a single object.
    if len(spec.output_content_paths) > 1 and len(spec.output_content_paths) != len(results):
        raise ExecFailureException(
            failure_type=FailureType.USER_FATAL,
            tip="Expected function to have %s outputs, but instead it had %s."
            % (len(spec.output_content_paths), len(results)),
        )

    if len(spec.output_content_paths) == 1:
        results = [results]

    inferred_result_types = [infer_artifact_type(res) for res in results]

    elapsedTime = timer.stop()
    _, peak = tracemalloc.get_traced_memory()
    system_metadata = {
        utils._RUNTIME_SEC_METRIC_NAME: str(elapsedTime),
        utils._MAX_MEMORY_MB_METRIC_NAME: str(peak / 10**6),
    }

    sys.path.pop(0)
    return results, inferred_result_types, system_metadata


def validate_spec(spec: FunctionSpec) -> None:
    if len(spec.input_content_paths) != len(spec.input_metadata_paths):
        raise Exception(
            "Found inconsistent number of input paths (%d) and input metadata paths (%d)"
            % (
                len(spec.input_content_paths),
                len(spec.input_metadata_paths),
            )
        )

    if len(spec.output_content_paths) != len(spec.output_metadata_paths):
        raise Exception(
            "Found inconsistent number of output paths (%d) and output metadata paths (%d)"
            % (
                len(spec.output_content_paths),
                len(spec.output_metadata_paths),
            )
        )
    if spec.expected_output_artifact_types is not None and len(
        spec.expected_output_artifact_types
    ) != len(spec.output_content_paths):
        raise Exception(
            "Found inconsistent number of expected output artifact types (%d) and output content paths (%d)"
            % (
                len(spec.expected_output_artifact_types),
                len(spec.output_content_paths),
            )
        )

    # Check and Metric operators must only have a single output.
    if (spec.operator_type == OperatorType.CHECK or spec.operator_type == OperatorType.METRIC) and (
        spec.expected_output_artifact_types is not None
        and len(spec.expected_output_artifact_types) != 1
    ):
        raise Exception("%s operators must only have a single output." % spec.operator_type)


def _cleanup(spec: FunctionSpec) -> None:
    """
    Cleans up any temporary files created during function execution.
    """
    # Delete the extracted fn file if it exists and the file path is not
    # something dangerous
    if spec.function_extract_path and spec.function_extract_path[-1] != "*":
        shutil.rmtree(spec.function_extract_path)


def run(spec: FunctionSpec) -> None:
    """
    Executes a function operator.
    """
    print("Started %s job: %s" % (spec.type, spec.name))

    exec_state = ExecutionState(user_logs=Logs())
    storage = parse_storage(spec.storage_config)
    try:
        validate_spec(spec)

        # Read the input data from intermediate storage.
        inputs, _, serialization_types = utils.read_artifacts(
            storage, spec.input_content_paths, spec.input_metadata_paths
        )

        derived_from_bson = SerializationType.BSON_TABLE in serialization_types
        print("Invoking the function...")
        results, result_types, system_metadata = _execute_function(spec, inputs, exec_state)
        if exec_state.status == ExecutionStatus.FAILED:
            # user failure
            utils.write_exec_state(storage, spec.metadata_path, exec_state)
            sys.exit(1)

        print("Function invoked successfully!")

        # Perform type checking on the function output.
        if spec.operator_type == OperatorType.METRIC:
            assert len(results) == 1, "Metric operator can only have a single output."
            result = results[0]

            if not (
                isinstance(result, int)
                or isinstance(result, float)
                or isinstance(result, np.number)
            ):
                raise ExecFailureException(
                    failure_type=FailureType.USER_FATAL,
                    tip=TIP_NOT_NUMERIC,
                )

        elif spec.operator_type == OperatorType.CHECK:
            assert len(results) == 1, "Check operator can only have a single output."
            check_result = results[0]

            if isinstance(check_result, pd.Series) and check_result.dtype == "bool":
                assert result_types[0] == ArtifactType.PICKLABLE

                # Cast pd.Series to a bool.
                # We only write True if every boolean in the series is True.
                series = pd.Series(check_result)
                check_passed = bool(series.size - series.sum().item() == 0)
            elif isinstance(check_result, bool) or isinstance(check_result, np.bool_):
                # Cast np.bool_ to a bool.
                check_passed = bool(check_result)
            else:
                raise ExecFailureException(
                    failure_type=FailureType.USER_FATAL,
                    tip=TIP_NOT_BOOL,
                )

            # If the check returned a value we interpret to mean 'false', we exit here, but
            # not before recording the output artifact value (which will be False).
            if not check_passed:
                print(f"Check Operator did not pass.")

                utils.write_artifact(
                    storage,
                    ArtifactType.BOOL,
                    derived_from_bson,  # derived_from_bson doesn't apply to bool artifact
                    spec.output_content_paths[0],
                    spec.output_metadata_paths[0],
                    check_passed,
                    system_metadata=system_metadata,
                )

                check_severity = spec.check_severity
                if spec.check_severity is None:
                    print(
                        "Check operator has an unspecified severity on spec. Defaulting to ERROR."
                    )
                    check_severity = CheckSeverity.ERROR

                failure_type = FailureType.USER_FATAL
                if check_severity == CheckSeverity.WARNING:
                    failure_type = FailureType.USER_NON_FATAL

                raise ExecFailureException(failure_type, tip=TIP_CHECK_DID_NOT_PASS)

            # If we get here, we know that the check has passed. The artifact type might need
            # still be updated. Eg. if the output was a pandas series.
            result_types[0] = ArtifactType.BOOL
            results[0] = True
        else:
            # Error if the number of expected outputs is not what was returned by the user's function.
            if len(results) != len(spec.expected_output_artifact_types):
                raise ExecFailureException(
                    failure_type=FailureType.USER_FATAL,
                    tip="Expected function to return %d outputs, but instead got %d."
                    % (len(spec.expected_output_artifact_types), len(results)),
                )

            for i, expected_output_type in enumerate(spec.expected_output_artifact_types):
                if (
                    expected_output_type != ArtifactType.UNTYPED
                    and expected_output_type != result_types[i]
                ):
                    raise ExecFailureException(
                        failure_type=FailureType.USER_FATAL,
                        tip="Expected type %s for the %d-th output of function, but it is of type %s."
                        % (expected_output_type, i, result_types[i]),
                    )

        for i, result in enumerate(results):
            utils.write_artifact(
                storage,
                result_types[i],
                derived_from_bson,
                spec.output_content_paths[i],
                spec.output_metadata_paths[i],
                result,
                system_metadata=system_metadata,
            )

        # If we made it here, then the operator has succeeded.
        exec_state.status = ExecutionStatus.SUCCEEDED
        print(f"Succeeded! Full logs: {exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)

    except ExecFailureException as e:
        # We must reconcile the user logs here, since those logs are not captured on the exception.
        from_exception_exec_state = ExecutionState.from_exception(e, user_logs=exec_state.user_logs)
        print(f"Failed with error. Full Logs:\n{from_exception_exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, from_exception_exec_state)
        sys.exit(1)

    except Exception as e:
        exec_state.mark_as_failure(
            FailureType.SYSTEM, TIP_UNKNOWN_ERROR, context=exception_traceback(e)
        )
        print(f"Failed with system error. Full Logs:\n{exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)
    finally:
        # Perform any cleanup
        _cleanup(spec)


def run_with_setup(spec: FunctionSpec) -> None:
    """
    Performs the setup needed for a Function operator and then executes it.
    """
    # Generate a unique function extract path if one does not exist already
    if not spec.function_extract_path:
        fn_extract_path = os.path.join(os.getcwd(), str(uuid.uuid4()))
        spec.function_extract_path = fn_extract_path

    op_path = get_extract_path.run(spec)

    extract_function.run(spec)

    requirements_path = os.path.join(op_path, "requirements.txt")
    if os.path.exists(requirements_path):
        os.system("{} -m pip install -r {}".format(sys.executable, requirements_path))

    run(spec)



import argparse
import base64

from aqueduct_executor.migrators.parameter_val_type_inference_000019 import execute
from aqueduct_executor.migrators.parameter_val_type_inference_000019.spec import parse_spec

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = parse_spec(spec_json)
    if spec.op == "encode":
        execute.run_type_inference_and_encode(spec.param_val)
    else:
        execute.run_decode(spec.param_val, spec.param_type)



import pandas as pd
from aqueduct_executor.operators.connectors.data import config, load, relational, utils
from aqueduct_executor.operators.utils.enums import ArtifactType
from sqlalchemy import create_engine, engine


class SqlServerConnector(relational.RelationalConnector):
    def __init__(self, config: config.SqlServerConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)

    def load(
        self, params: load.RelationalParams, df: pd.DataFrame, artifact_type: ArtifactType
    ) -> None:
        if artifact_type != ArtifactType.TABLE:
            raise Exception("The data being loaded must be of type table, found %s" % artifact_type)
        # NOTE (saurav): PyODBC for SQL Server does not support `method="multi"` for `df.to_sql`,
        # which is why SqlServerConnector overrides `load`.
        df.to_sql(
            params.table,
            con=self.engine,
            if_exists="replace",
            index=False,
        )


def _create_engine(config: config.SqlServerConfig) -> engine.Engine:
    # SQL Server Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/mssql.html#dialect-mssql-pyodbc-connect
    url = "mssql+pyodbc://{username}:{password}@{host}:{port}/{database}?driver=ODBC+Driver+17+for+SQL+Server".format(
        username=config.username,
        password=utils.url_encode(config.password),
        host=config.host,
        port=config.port,
        database=config.database,
    )

    # We use `fast_executemany=True` to improve the performance of writing a large DataFrame.
    # https://docs.sqlalchemy.org/en/14/changelog/migration_13.html#support-for-pyodbc-fast-executemany
    return create_engine(url, fast_executemany=True)



import sys

from aqueduct.utils.serialization import deserialize
from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.operators.param_executor.spec import ParamSpec
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.enums import ExecutionStatus, FailureType
from aqueduct_executor.operators.utils.execution import (
    TIP_UNKNOWN_ERROR,
    Error,
    ExecFailureException,
    ExecutionState,
    Logs,
    exception_traceback,
)
from aqueduct_executor.operators.utils.storage.parse import parse_storage


def run(spec: ParamSpec) -> None:
    """Parameter operators are unique in that the output content is expected to have already been populated correctly.

    Therefore, this operator is responsible only for:
    - Checking that the parameter type matches the expected type.
    - Writing the operator and artifact metadata to storage, so that orchestration can proceed as normal.

    The artifact output paths are written to before any type checking occurs, so all artifact-related paths are
    expected to have been populated, even if the operator itself fails. However, this is not guaranteed to be the case,
    since system errors are still possible.
    """
    print("Job Spec: \n{}".format(spec.json()))

    storage = parse_storage(spec.storage_config)

    try:
        val = deserialize(
            spec.serialization_type,
            spec.expected_type,
            storage.get(spec.output_content_path),
        )
        inferred_type = infer_artifact_type(val)

        # This does not write to the output artifact's content path as a performance optimization.
        # That has already been written by the Golang Orchestrator.
        utils.write_artifact(
            storage,
            inferred_type,
            False,  # derived_from_bson
            None,  # output_content_path
            spec.output_metadata_path,
            val,
            system_metadata={},
        )

        if inferred_type != spec.expected_type:
            raise ExecFailureException(
                failure_type=FailureType.USER_FATAL,
                tip="Supplied parameter expects type `%s`, but got `%s` instead."
                % (spec.expected_type, inferred_type),
            )

        utils.write_exec_state(
            storage,
            spec.metadata_path,
            ExecutionState(status=ExecutionStatus.SUCCEEDED, user_logs=Logs()),
        )

    except ExecFailureException as e:
        from_exception_exec_state = ExecutionState.from_exception(e, user_logs=Logs())
        print(f"Failed with error. Full Logs:\n{from_exception_exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, from_exception_exec_state)
        sys.exit(1)
    except Exception as e:
        exec_state = ExecutionState(
            status=ExecutionStatus.FAILED,
            failure_type=FailureType.SYSTEM,
            error=Error(context=exception_traceback(e), tip=TIP_UNKNOWN_ERROR),
            user_logs=Logs(),
        )
        print(f"Failed with system error. Full Logs:\n{exec_state.json()}")
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)



from aqueduct_executor.operators.connectors.data import config, relational, utils
from sqlalchemy import create_engine, engine


class SnowflakeConnector(relational.RelationalConnector):
    def __init__(self, config: config.SnowflakeConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)


def _create_engine(config: config.SnowflakeConfig) -> engine.Engine:
    # Snowflake Dialect:
    # https://github.com/snowflakedb/snowflake-sqlalchemy
    url = "snowflake://{username}:{password}@{account_identifier}/{database}/{schema}?warehouse={warehouse}".format(
        username=config.username,
        password=utils.url_encode(config.password),
        account_identifier=config.account_identifier,
        database=config.database,
        schema=config.db_schema,
        warehouse=config.warehouse,
    )
    return create_engine(url)



from aqueduct_executor.operators.connectors.data import config, relational, utils
from sqlalchemy import create_engine, engine


class MySqlConnector(relational.RelationalConnector):
    def __init__(self, config: config.MySqlConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)


def _create_engine(config: config.MySqlConfig) -> engine.Engine:
    # MySQL Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/mysql.html#module-sqlalchemy.dialects.mysql.mysqldb
    url = "mysql+mysqldb://{username}:{password}@{host}:{port}/{database}".format(
        username=config.username,
        password=utils.url_encode(config.password),
        host=config.host,
        port=config.port,
        database=config.database,
    )
    return create_engine(url)



from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from aqueduct_executor.operators.connectors.data import connector, extract, load
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from sqlalchemy import MetaData, engine, inspect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base


class RelationalConnector(connector.DataConnector):
    def __init__(self, conn_engine: engine.Engine):
        self.engine = conn_engine

    def __del__(self) -> None:
        self.engine.dispose()

    def authenticate(self) -> None:
        try:
            self.engine.connect()
        except SQLAlchemyError as e:
            raise ConnectionError("Unable to connect.") from e

    def discover(self) -> List[str]:
        return inspect(self.engine).get_table_names()  # type: ignore

    def extract(self, params: extract.RelationalParams) -> Any:
        assert params.usable(), "Query is not usable. Did you forget to expand placeholders?"
        return pd.read_sql(params.query, con=self.engine)

    def _delete_object(self, name: str, context: Optional[Dict[str, Any]] = None) -> None:
        if context:
            metadata = context["metadata"]
            base = context["base"]
        else:
            raise Exception("Unexpectedly cannot find context for deletion.")
        sql_table = metadata.tables[name]
        base.metadata.drop_all(self.engine, [sql_table], checkfirst=True)

    def delete(self, tables: List[str]) -> List[SavedObjectDelete]:
        results = []
        base = declarative_base()
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        delete_helper: Callable[[str], None] = lambda name: self._delete_object(
            name, context={"metadata": metadata, "base": base}
        )
        for table in tables:
            results.append(
                delete_object(
                    table,
                    delete_helper,
                )
            )
        return results

    def load(
        self, params: load.RelationalParams, df: pd.DataFrame, artifact_type: ArtifactType
    ) -> None:
        if artifact_type != ArtifactType.TABLE:
            raise Exception("The data being loaded must be of type table, found %s" % artifact_type)
        # NOTE (saurav): df._to_sql has known performance issues. Using `method="multi"` helps incrementally,
        # since pandas will pass multiple rows in a single INSERT. If this still remains an issue, we can pass in a
        # callable function for `method` that does bulk loading.
        # See: https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method
        df.to_sql(
            params.table,
            con=self.engine,
            if_exists=params.update_mode.value,
            index=False,
            method="multi",
        )



import argparse
import base64
import json

from aqueduct_executor.operators.function_executor.spec import FunctionSpec, parse_spec


def run(spec: FunctionSpec) -> str:
    return spec.function_extract_path


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    try:
        spec = parse_spec(spec_json)
    except Exception:
        # Log the offending spec on failures.
        print("Failing raw spec: ", args.spec)
        raise

    # The output of the print statement to stdout is captured by the calling bash script into a variable,
    # so we should not include any other happy-path print statements in this Python script.
    print(run(spec))



from aqueduct_executor.operators.utils.enums import FailureType


class MissingConnectorDependencyException(Exception):
    """Exception raised due to the connector integration's dependencies aren't installed."""

    pass


class MissingInputPathsException(Exception):
    """Exception raised due to input data not being supplied."""

    pass



import sqlite3

import pandas as pd
from aqueduct_executor.operators.connectors.data import config, load, relational
from aqueduct_executor.operators.utils.enums import ArtifactType
from packaging.version import parse as parse_version
from sqlalchemy import create_engine, engine

# https://www.sqlite.org/limits.html#max_variable_number
SQLITE_MAX_VARIABLE_NUMBER = 32766
SQLITE_MAX_VARIABLE_NUMBER_EARLIER_VERSION = 999


class SqliteConnector(relational.RelationalConnector):
    def __init__(self, config: config.SqliteConfig):
        if parse_version(sqlite3.sqlite_version) >= parse_version("3.32.0"):
            self.variable_number_limit = SQLITE_MAX_VARIABLE_NUMBER
        else:
            self.variable_number_limit = SQLITE_MAX_VARIABLE_NUMBER_EARLIER_VERSION

        conn_engine = _create_engine(config)
        super().__init__(conn_engine)

    def load(
        self, params: load.RelationalParams, df: pd.DataFrame, artifact_type: ArtifactType
    ) -> None:
        if artifact_type != ArtifactType.TABLE:
            raise Exception("The data being loaded must be of type table, found %s" % artifact_type)
        # NOTE (saurav): df._to_sql has known performance issues. Using `method="multi"` helps incrementally,
        # since pandas will pass multiple rows in a single INSERT. If this still remains an issue, we can pass in a
        # callable function for `method` that does bulk loading.
        # See: https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method
        df.to_sql(
            params.table,
            con=self.engine,
            if_exists=params.update_mode.value,
            index=False,
            method="multi",
            # We need to specify chunksize due to sqlite3's variable number limit.
            # chunksize corresponds to the max number of rows in each batch to be written at a time.
            # Variable number is the multiplication of the row count and the column count, so we can
            # calculate chunksize by dividing the max variable number with the number of columns.
            chunksize=int(self.variable_number_limit / len(df.columns)),
        )


def _create_engine(config: config.SqliteConfig) -> engine.Engine:
    # SQLite Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#dialect-sqlite-pysqlite-connect
    url = "sqlite:///{database}".format(
        database=config.database,
    )
    return create_engine(url)



from typing import Any, Dict, List, Optional

import awswrangler as wr
import pandas as pd
from aqueduct_executor.operators.connectors.data import connector, extract, load
from aqueduct_executor.operators.connectors.data.config import AthenaConfig
from aqueduct_executor.operators.connectors.data.utils import construct_boto_session
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete

DEFAULT_CATALOG = "AwsDataCatalog"
LIST_TABLES_QUERY_ATHENA = "AQUEDUCT_ATHENA_LIST_TABLE"


class AthenaConnector(connector.DataConnector):
    def __init__(self, config: AthenaConfig):
        self.session = construct_boto_session(config)
        self.output_location = config.output_location
        self.database = config.database

    def _list_tables(self) -> List[str]:
        client = self.session.client("athena")
        tables = client.list_table_metadata(CatalogName=DEFAULT_CATALOG, DatabaseName=self.database)
        return [table["Name"] for table in tables["TableMetadataList"]]

    def authenticate(self) -> None:
        self._list_tables()
        # This checks whether the S3 output path is valid.
        wr.athena.read_sql_query(
            sql="SELECT 1;",
            database=self.database,
            boto3_session=self.session,
            ctas_approach=False,
            s3_output=self.output_location,
            keep_files=False,
        )

    def discover(self) -> List[str]:
        return self._list_tables()

    def extract(self, params: extract.RelationalParams) -> pd.DataFrame:
        assert params.usable(), "Query is not usable. Did you forget to expand placeholders?"
        if params.query == LIST_TABLES_QUERY_ATHENA:
            return pd.DataFrame(self._list_tables(), columns=["Tables"])
        else:
            return wr.athena.read_sql_query(
                sql=params.query,
                database=self.database,
                boto3_session=self.session,
                # Disabling ctas improves generality at the cost of performance.
                # More details here: https://aws-sdk-pandas.readthedocs.io/en/stable/stubs/awswrangler.athena.read_sql_query.html
                ctas_approach=False,
                s3_output=self.output_location,
                keep_files=False,
            )

    def load(
        self, params: load.RelationalParams, df: pd.DataFrame, artifact_type: ArtifactType
    ) -> None:
        raise Exception("Save operation not supported for Athena.")

    def delete(self, objects: List[str]) -> List[SavedObjectDelete]:
        raise Exception("Delete operation not supported for Athena.")



from aqueduct_executor.operators.utils.storage.config import FileStorageConfig
from aqueduct_executor.operators.utils.storage.storage import Storage


class FileStorage(Storage):
    _config: FileStorageConfig

    def __init__(self, config: FileStorageConfig):
        self._config = config

    def put(self, key: str, value: bytes) -> None:
        path = self.get_full_path(key)
        print(f"writing to file: {path}")
        with open(path, "wb") as f:
            f.write(value)

    def get(self, key: str) -> bytes:
        path = self.get_full_path(key)
        print(f"reading from file: {path}")
        with open(path, "rb") as f:
            return f.read()

    def get_full_path(self, key: str) -> str:
        return self._config.directory + "/" + key



import io
import json
from typing import Any, Dict, List, Optional

import cloudpickle as pickle
import numpy as np
import pandas as pd
from aqueduct_executor.operators.connectors.data import common, connector, extract, load
from aqueduct_executor.operators.connectors.data.config import S3Config
from aqueduct_executor.operators.connectors.data.utils import construct_boto_session
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from PIL import Image

_DEFAULT_JSON_ENCODING = "utf8"
_DEFAULT_IMAGE_FORMAT = "jpeg"


class S3Connector(connector.DataConnector):
    def __init__(self, config: S3Config):
        session = construct_boto_session(config)
        self.s3 = session.resource("s3")
        self.bucket = config.bucket

    def authenticate(self) -> None:
        bucket = self.s3.Bucket(self.bucket)
        # Below is a low-overhead way of checking if the user has access to the bucket.
        # Source: https://stackoverflow.com/a/49817544
        if not bucket.creation_date:
            raise Exception(
                "Bucket does not exist or you do not have permission to access the bucket."
            )

    def discover(self) -> List[str]:
        raise Exception("Discover is not supported for S3.")

    def _fetch_object(self, key: str, params: extract.S3Params) -> Any:
        response = self.s3.Object(self.bucket, key).get()
        data = response["Body"].read()
        if params.artifact_type == ArtifactType.TABLE:
            if params.format is None:
                raise Exception("You must specify a file format for table data.")
            buf = io.BytesIO(data)
            if params.format == common.S3TableFormat.CSV:
                return pd.read_csv(buf)
            elif params.format == common.S3TableFormat.JSON:
                return pd.read_json(buf)
            elif params.format == common.S3TableFormat.PARQUET:
                return pd.read_parquet(buf)
            raise Exception("Unknown S3 file format %s." % params.format)
        elif params.artifact_type == ArtifactType.JSON:
            # This assumes that the encoding is "utf-8". May worth considering letting the user
            # specify custom encoding in the future.
            json_data = data.decode(_DEFAULT_JSON_ENCODING)
            # Make sure the data is a valid json object.
            try:
                json.loads(json_data)
                return json_data
            except:
                raise Exception("The file is not a valid JSON object.")
        elif params.artifact_type == ArtifactType.IMAGE:
            return Image.open(io.BytesIO(data))
        elif params.artifact_type == ArtifactType.BYTES:
            return data
        elif (
            params.artifact_type == ArtifactType.STRING
            or params.artifact_type == ArtifactType.BOOL
            or params.artifact_type == ArtifactType.NUMERIC
            or params.artifact_type == ArtifactType.DICT
            or params.artifact_type == ArtifactType.TUPLE
            or params.artifact_type == ArtifactType.PICKLABLE
        ):
            unpickled_data = pickle.loads(data)

            if params.artifact_type == ArtifactType.STRING:
                if not isinstance(unpickled_data, str):
                    raise Exception(
                        "The file is expected to be a string, got %s." % type(unpickled_data)
                    )
            elif params.artifact_type == ArtifactType.BOOL:
                if not (isinstance(unpickled_data, bool) or isinstance(unpickled_data, np.bool_)):
                    raise Exception(
                        "The file is expected to be a bool, got %s." % type(unpickled_data)
                    )
            elif params.artifact_type == ArtifactType.NUMERIC:
                if not (
                    isinstance(unpickled_data, int)
                    or isinstance(unpickled_data, float)
                    or isinstance(unpickled_data, np.number)
                ):
                    raise Exception(
                        "The file is expected to be a numeric, got %s." % type(unpickled_data)
                    )
            elif params.artifact_type == ArtifactType.DICT:
                if not isinstance(unpickled_data, dict):
                    raise Exception(
                        "The file is expected to be a dictionary, got %s." % type(unpickled_data)
                    )
            elif params.artifact_type == ArtifactType.TUPLE:
                if not isinstance(unpickled_data, tuple):
                    raise Exception(
                        "The file is expected to be a tuple, got %s." % type(unpickled_data)
                    )

            return unpickled_data
        else:
            raise Exception("Unsupported data type %s." % params.artifact_type)

    def extract(self, params: extract.S3Params) -> Any:
        path = json.loads(params.filepath)
        if not isinstance(path, List):
            if len(path) == 0:
                raise Exception("S3 file path cannot be an empty string.")
            if path[-1] == "/":
                # This means the path is a directory, and we will do a prefix search.
                files = []
                for obj in self.s3.Bucket(self.bucket).objects.filter(Prefix=path):
                    # The filter api also returns the directories, so we filter them out.
                    if (obj.key)[-1] != "/":
                        files.append(self._fetch_object(obj.key, params))

                if params.artifact_type == ArtifactType.TABLE and params.merge:
                    return pd.concat(files)
                else:
                    return tuple(files)
            else:
                # This means the path is a file name, and we do a regular file retrieval.
                return self._fetch_object(path, params)
        else:
            # This means we have a list of file paths.
            files = []
            for key in path:
                if len(key) == 0:
                    raise Exception("S3 file path cannot be an empty string.")
                if key[-1] == "/":
                    raise Exception("Each key in the list must not be a directory, found %s." % key)
                files.append(self._fetch_object(key, params))

            if params.artifact_type == ArtifactType.TABLE and params.merge:
                return pd.concat(files)
            else:
                return tuple(files)

    def load(self, params: load.S3Params, data: Any, artifact_type: ArtifactType) -> None:
        if artifact_type == ArtifactType.TABLE:
            if params.format is None:
                raise Exception("You must specify a file format for table data.")
            buf = io.BytesIO(data)
            if params.format == common.S3TableFormat.CSV:
                data.to_csv(buf, index=False)
            elif params.format == common.S3TableFormat.JSON:
                # Index cannot be False for `to.json` for default orient
                # See: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_json.html
                data.to_json(buf)
            elif params.format == common.S3TableFormat.PARQUET:
                data.to_parquet(buf, index=False)
            else:
                raise Exception("Unknown S3 file format %s." % params.format)
            serialized_data = buf.getvalue()
        elif artifact_type == ArtifactType.JSON:
            serialized_data = data.encode(_DEFAULT_JSON_ENCODING)
        elif artifact_type == ArtifactType.IMAGE:
            img_bytes = io.BytesIO()
            data.save(img_bytes, format=_DEFAULT_IMAGE_FORMAT)
            serialized_data = img_bytes.getvalue()
        elif artifact_type == ArtifactType.BYTES:
            serialized_data = data
        elif (
            artifact_type == ArtifactType.STRING
            or artifact_type == ArtifactType.BOOL
            or artifact_type == ArtifactType.NUMERIC
            or artifact_type == ArtifactType.DICT
            or artifact_type == ArtifactType.TUPLE
            or artifact_type == ArtifactType.PICKLABLE
        ):
            serialized_data = pickle.dumps(data)
        else:
            raise Exception("Unsupported data type %s." % artifact_type)

        self.s3.Object(self.bucket, params.filepath).put(Body=serialized_data)

    def _delete_object(self, name: str, context: Optional[Dict[str, Any]] = None) -> None:
        self.s3.Object(self.bucket, name).delete()

    def delete(self, objects: List[str]) -> List[SavedObjectDelete]:
        results = []
        for key in objects:
            results.append(delete_object(key, self._delete_object))
        return results



from abc import ABC, abstractmethod
from typing import Any, List

from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete


class DataConnector(ABC):
    @abstractmethod
    def authenticate(self) -> None:
        """Authenticates connector configuration. Raises a ConnectionError if there is an error."""

    @abstractmethod
    def discover(self) -> List[str]:
        """Discover items in the connection.

        Returns:
            A list of items discovered.
        """

    @abstractmethod
    def extract(  # type: ignore
        self,
        # TODO (ENG-1285): Revisit the typing issue that araises from inheritence
        params,  # extract.Params
    ) -> Any:
        """Extracts data from source into a DataFrame.

        Args:
            params: Extract parameters for the connector.

        Returns:
            A DataFrame that contains the data extracted by the connector.
        """

    @abstractmethod
    def delete(  # type: ignore
        self,
        # TODO (ENG-1285): Revisit the typing issue that araises from inheritence
        objects,  # List[str]
    ) -> List[SavedObjectDelete]:
        """Delete objects from integration.

        Args:
            objects: List of objects to delete from this integration.
        """

    @abstractmethod
    def load(  # type: ignore
        self,
        # TODO (ENG-1285): Revisit the typing issue that araises from inheritence
        params,  # load.Params
        data: Any,
        artifact_type: ArtifactType,
    ) -> None:
        """Loads data into destination storage integration.

        Args:
            params: Load parameters for the connector.
            data: data to load.
            artifact_type: type of the artifact.
        """



import os
from pathlib import Path

from setuptools import find_packages, setup

install_requires = open("requirements.txt").read().strip().split("\n")

readme_path = Path(os.environ["PWD"], "../../README.md")
long_description = open(readme_path).read()

setup(
    name="aqueduct-ml",
    version="0.1.8",
    install_requires=install_requires,
    scripts=["bin/aqueduct"],
    packages=find_packages(),
    description="Prediction Infrastructure for Data Scientists",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.aqueducthq.com/",
    license="Apache License 2.0",
    author="Aqueduct, Inc.",
    author_email="hello@aqueducthq.com",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
)



import json
import re
import uuid
from datetime import date
from typing import Any, Dict, List, Optional, Union

from aqueduct.integrations.sql_integration import BUILT_IN_EXPANSIONS, PREV_TABLE_TAG, TAG_PATTERN
from aqueduct_executor.operators.connectors.data import common, models
from aqueduct_executor.operators.utils.enums import ArtifactType
from pydantic import parse_obj_as


def replace_today() -> str:
    return "'" + date.today().strftime("%Y-%m-%d") + "'"


def _expand_placeholders(query: str, parameters: Dict[str, str]) -> str:
    """Expands any tags found in the raw query, eg. {{ today }}.

    Relational queries can be arbitrarily parameterized the same way operators are. The only
    requirement is that these parameters must be defined as strings.

    User-defined parameters are prioritized over built-in ones. Eg. if the user defines a parameter
    named "today" that they set with value "1234", the "{{today}}" will be expanded as "1234", even
    though there already is a built-in expansion.
    """
    matches = re.findall(TAG_PATTERN, query)
    for match in matches:
        tag_name = match.strip(" {}")

        if tag_name in parameters:
            query = query.replace(match, parameters[tag_name])
        elif tag_name in BUILT_IN_EXPANSIONS:
            expansion_func = BUILT_IN_EXPANSIONS[tag_name]
            query = query.replace(match, expansion_func())
        else:
            # If there's a tag in the query for which no expansion value is available, we error here.
            raise Exception("Unable to expand tag `%s` for query `%s`." % (tag_name, query))
    return query


class RelationalParams(models.BaseParams):
    # The query cannot be used until `apply_placeholders()` is called on it. This flushes out
    # any user-defined tags like `{{today}}`.
    query_is_usable: Optional[bool] = False

    # Exactly one of 'query' and 'queries' will be set.
    # `query` represents a single query.
    query: Optional[str] = None
    # `queries` represents a chain of queries. We must first run _compile_chain
    # to compile it to a single query before further processing.
    queries: Optional[List[str]] = None

    # TODO: Consider not including github as part of relational params when it is JSON marshalled
    github_metadata: Optional[Any]

    def _compile_chain(self, queries: List[str]) -> str:
        """
        `_compile_chain` compiles a chain query to a single query using `WITH` clause.
        We generate temp_table_name and replace PREV_TABLE_TAG accordingly.

        Example:
        queries: [
            "SELECT * FROM my_table",
            "SELECT field_a, field_b FROM $",
            "SELECT * FROM $",
        ]

        returns: `
            WITH
                generated_tmp_a AS (SELECT * FROM my_table),
                generated_tmp_b AS (SELECT field_a, field_b FROM generated_tmp_a)
            SELECT * FROM generated_tmp_b
        `
        """
        if not queries:
            return ""

        if len(queries) == 1:
            return queries[0]

        with_clause = "WITH\n"
        prev_table_name = ""
        normalized_query = ""
        for (idx, query) in enumerate(queries):
            # Remove spaces and trailing semicolumns if any.
            normalized_query = query.strip().rstrip(";")

            # Replace tag except for the first query
            if idx == 0:
                if PREV_TABLE_TAG in normalized_query:
                    raise Exception(
                        f"Cannot compile chain. {PREV_TABLE_TAG} appears in the first query: {query}"
                    )
            else:
                normalized_query = normalized_query.replace(PREV_TABLE_TAG, prev_table_name)

            # Subquery goes to the 'WITH' clause except for the last one.
            if idx < len(queries) - 1:
                cur_table_name = f"aqueduct_{uuid.uuid4().hex}"
                with_clause += f"{cur_table_name} AS (\n{normalized_query}\n)"

                # There are more subqueries to append in this 'WITH' clause
                if idx < len(queries) - 2:
                    with_clause += ",\n"
                prev_table_name = cur_table_name

        # Returns `WITH` clause with the normalized final query.
        return f"{with_clause}\n{normalized_query}"

    def compile(self, parameters: Dict[str, str]) -> None:
        """
        `compile` compiles this object to a single query that can be
        executed.
        """
        assert (
            int(bool(self.query)) + int(bool(self.queries)) == 1
        ), "Exactly one of .query and .queries fields should be set."

        queries = self.queries or []
        if self.query:
            queries = [self.query]

        print(f"Compiling queries {queries} .")
        query = self._compile_chain(queries)
        print(f"Compiled query is {query} .")

        query = _expand_placeholders(query, parameters)
        print(f"Expanded query is `{query}`.")
        self.query = query
        self.query_is_usable = True

    def usable(self) -> bool:
        """Denotes whether all placeholders have already been expanded for this query.

        Callers should check that `usable()` -> True before actually executing this query.
        """
        # We cannot return self.query_is_usable directly, since it is an Optional
        # and the method expects a bool to be returned.
        return bool(self.query_is_usable)


class S3Params(models.BaseParams):
    filepath: str
    artifact_type: ArtifactType
    format: Optional[common.S3TableFormat]
    merge: Optional[bool]


class MongoDBFindParams(models.BaseParams):
    args: Optional[List[Any]] = None
    kwargs: Optional[Dict[str, Any]] = None


class MongoDBParams(models.BaseParams):
    collection: str
    query_serialized: str
    query: Optional[MongoDBFindParams] = None

    def compile(self, parameters: Dict[str, str]) -> None:
        expanded = _expand_placeholders(self.query_serialized, parameters)
        self.query = parse_obj_as(MongoDBFindParams, json.loads(expanded))

    def usable(self) -> bool:
        return bool(self.query) and bool(self.collection)


Params = Union[RelationalParams, S3Params, MongoDBParams]



import base64
import json

from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.migrators.parameter_val_type_inference_000019 import serialize
from aqueduct_executor.operators.utils.enums import SerializationType


def run_type_inference_and_encode(json_val: str) -> None:
    """
    Infers the type for a given param value and print to stdout
    Encode to base64 and print out as well
    """
    val = json.loads(json_val)
    artifact_type = infer_artifact_type(val)
    serialization_type = serialize.artifact_type_to_serialization_type(artifact_type, val)
    print(serialization_type.value)
    print(serialize.serialize_val(val, serialization_type))


def run_decode(val: str, val_type: str) -> None:
    """
    Decodes from base64 and prints out the value
    """
    decoded_val = base64.b64decode(val)
    serialization_type = SerializationType(val_type)
    deserialized_val = serialize.deserialization_function_mapping[serialization_type](decoded_val)
    print(json.dumps(deserialized_val))



import json
from typing import List

from pydantic import BaseModel, parse_obj_as

try:
    from typing import Literal
except ImportError:
    # Python 3.7 does not support typing.Literal
    from typing_extensions import Literal  # type: ignore

from aqueduct_executor.operators.utils import enums
from aqueduct_executor.operators.utils.storage import config


class SystemMetricSpec(BaseModel):
    name: str
    type: Literal[enums.JobType.SYSTEM_METRIC]
    storage_config: config.StorageConfig
    metadata_path: str
    metric_name: str
    input_metadata_paths: List[str]
    output_content_path: str
    output_metadata_path: str


def parse_spec(spec_json: bytes) -> SystemMetricSpec:
    """
    Parses a JSON string into a FunctionSpec.
    """
    data = json.loads(spec_json)
    return parse_obj_as(SystemMetricSpec, data)



import argparse
import base64
import io
import os
import shutil
import sys
import traceback
import zipfile

from aqueduct_executor.operators.function_executor.spec import FunctionSpec, parse_spec
from aqueduct_executor.operators.function_executor.utils import OP_DIR
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.storage.storage import Storage


def _unzip_function_contents(function_in_bytes: bytes, extract_path: str) -> None:
    """
    Unzips raw function bytes into the current directory, assuming the bytes represent
    a zip file.
    """
    with zipfile.ZipFile(io.BytesIO(function_in_bytes), "r") as z:
        toplevel_dir = _extract_folder_name(z)
        for name in z.namelist():
            z.extract(name, extract_path)

        # Assumes all extracted file will be located in nested path.
        nested_path = os.path.join(extract_path, toplevel_dir)
        op_path = os.path.join(extract_path, OP_DIR)
        shutil.rmtree(op_path, ignore_errors=True)
        shutil.copytree(nested_path, op_path)
        shutil.rmtree(nested_path)


def _extract_folder_name(zip_ref: zipfile.ZipFile) -> str:
    """
    Given a zip file, return the name of the top-level folder.

    Assumption: The first item in namelist() is typically the name of the folder in
    which the files are extracted into. This is the same assumption `generate_foldername.sh`
    operates on.
    """
    folders = [folder for folder in zip_ref.namelist() if folder.endswith("/")]

    if len(folders) == 0:
        raise Exception("No folders found in zip file.")
    return folders[0]


def extract_function(storage: Storage, spec: FunctionSpec) -> None:
    """
    Extracts the user-specified function.
    """
    fn_path = spec.function_extract_path
    if not os.path.exists(fn_path):
        os.makedirs(fn_path)

    function_byte = storage.get(spec.function_path)
    _unzip_function_contents(
        function_in_bytes=function_byte,
        extract_path=fn_path,
    )


def run(spec: FunctionSpec) -> None:
    print("Job Spec: \n{}".format(spec.json()))

    try:
        storage = parse_storage(spec.storage_config)
        extract_function(storage, spec)
    except Exception as e:  # Catch all error types
        print("Exception Raised: ", e)
        traceback.print_tb(e.__traceback__)
        sys.exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = parse_spec(spec_json)

    run(spec)



from typing import Any, List, Optional

import pandas as pd
from aqueduct_executor.operators.connectors.data import common, connector, extract, load
from aqueduct_executor.operators.connectors.data.config import MongoDBConfig
from aqueduct_executor.operators.utils.enums import ArtifactType
from aqueduct_executor.operators.utils.saved_object_delete import SavedObjectDelete
from aqueduct_executor.operators.utils.utils import delete_object
from pymongo import MongoClient
from pymongo.client_session import ClientSession
from pymongo.database import Database


class MongoDBConnector(connector.DataConnector):
    _client: MongoClient[Any]
    _db_name: str

    def __init__(self, config: MongoDBConfig):
        self._client = MongoClient(
            config.auth_uri,
            connect=True,
        )
        self._db_name = config.database
        self._test()

    def _test(self) -> None:
        try:
            self._client.test
            assert (
                self._db_name in self._client.list_database_names()
            ), f"Database {self._db_name} does not exist."
        except Exception as e:
            raise ConnectionError("Unable to connect") from e

    def __del__(self) -> None:
        self._client.close()

    def _connect_db(self, session: Optional[ClientSession] = None) -> Database[Any]:
        if session:
            return session.client[self._db_name]
        return self._client[self._db_name]

    def authenticate(self) -> None:
        self._test()

    def _discover(self, session: Optional[ClientSession] = None) -> List[str]:
        return self._connect_db(session).list_collection_names()

    def discover(self) -> List[str]:
        return self._discover()

    def extract(self, params: extract.MongoDBParams) -> Any:
        assert params.usable()
        query = params.query
        assert query is not None

        db = self._connect_db()
        collection = params.collection
        assert (
            collection in db.list_collection_names()
        ), f"Collection `{collection}` does not exist."

        raw_results = db[collection].find(*(query.args or []), **(query.kwargs or {}))
        return pd.DataFrame(raw_results)

    def load(
        self, params: load.RelationalParams, df: pd.DataFrame, artifact_type: ArtifactType
    ) -> None:
        if artifact_type != ArtifactType.TABLE:
            raise Exception("The data being loaded must be of type table, found %s" % artifact_type)

        with self._client.start_session() as session:
            with session.start_transaction():
                db = self._connect_db(session)
                collections = self._discover(session)
                exists = params.table in collections
                collection = None
                replace = False
                if exists:
                    if params.update_mode == common.UpdateMode.FAIL:
                        raise Exception(f"Specified collection {params.table} already exists.")

                    collection = db[params.table]
                    if params.update_mode == common.UpdateMode.REPLACE:
                        replace = True
                else:
                    collection = db.create_collection(params.table)

                if replace:
                    collection.delete_many({})
                collection.insert_many(df.to_dict("records"))

    def delete(self, tables: List[str]) -> List[SavedObjectDelete]:
        results = []
        db = self._connect_db()

        # This helper simply bypass type check of `delete_object`,
        # which requires the callback to return None.
        def _delete_table(t: str) -> None:
            db.drop_collection(t)

        for table in tables:
            results.append(delete_object(table, lambda t: _delete_table(t)))
        return results



import base64
import io
import json
import os
import shutil
import tempfile
import uuid
from pathlib import Path
from typing import Any, Callable, Dict, cast

import cloudpickle as pickle
import pandas as pd
from aqueduct_executor.operators.utils.enums import ArtifactType, SerializationType
from PIL import Image

_DEFAULT_ENCODING = "utf8"

# The temporary file name that a Tensorflow keras model will be dumped into before we read/write it from storage.
# This will be cleaned up within the serialization logic.
_TEMP_KERAS_MODEL_NAME = "keras_model"


def _read_table_content(content: bytes) -> pd.DataFrame:
    return pd.read_json(io.BytesIO(content), orient="table")


def _read_json_content(content: bytes) -> Any:
    return json.loads(content.decode(_DEFAULT_ENCODING))


def _read_pickle_content(content: bytes) -> Any:
    return pickle.loads(content)


def _read_image_content(content: bytes) -> Image.Image:
    return Image.open(io.BytesIO(content))


def _read_string_content(content: bytes) -> str:
    return content.decode(_DEFAULT_ENCODING)


def _read_bytes_content(content: bytes) -> bytes:
    return content


def make_temp_dir() -> str:
    """
    Create a unique, temporary directory in the local filesystem and returns the path.
    """
    dir_path = None
    created = False
    # Try to create the directory. If it already exists, try again with a new name.
    while not created:
        dir_path = Path(tempfile.gettempdir()) / str(uuid.uuid4())
        try:
            os.mkdir(dir_path)
            created = True
        except FileExistsError:
            pass

    assert dir_path is not None
    return str(dir_path)


# Returns a tf.keras.Model type. We don't assume that every user has it installed,
# so we return "Any" type.
def _read_tf_keras_model(content: bytes) -> Any:
    temp_model_dir = None
    try:
        temp_model_dir = make_temp_dir()
        model_file_path = os.path.join(temp_model_dir, _TEMP_KERAS_MODEL_NAME)
        with open(model_file_path, "wb") as f:
            f.write(content)

        from tensorflow import keras

        return keras.load_model(model_file_path)
    finally:
        if temp_model_dir is not None and os.path.exists(temp_model_dir):
            shutil.rmtree(temp_model_dir)


deserialization_function_mapping: Dict[str, Callable[[bytes], Any]] = {
    SerializationType.TABLE: _read_table_content,
    SerializationType.JSON: _read_json_content,
    SerializationType.PICKLE: _read_pickle_content,
    SerializationType.IMAGE: _read_image_content,
    SerializationType.STRING: _read_string_content,
    SerializationType.BYTES: _read_bytes_content,
    SerializationType.TF_KERAS: _read_tf_keras_model,
}



import json
from typing import Any

import cloudpickle as pickle
from aqueduct.utils.type_inference import infer_artifact_type
from aqueduct_executor.migrators.artifact_migration_000016.spec import MigrationSpec
from aqueduct_executor.operators.utils.enums import ArtifactType, SerializationType
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.storage.storage import Storage

# The variable definition and type mapping logic below are partially
# duplicated from aqueduct_executor.operators.utils.utils

_DEFAULT_ENCODING = "utf8"
_METADATA_ARTIFACT_TYPE_KEY = "artifact_type"
_METADATA_SERIALIZATION_TYPE_KEY = "serialization_type"


def _write_string_output(
    storage: Storage,
    output_path: str,
    output: str,
) -> None:
    storage.put(output_path, output.encode(_DEFAULT_ENCODING))


def _write_pickle_output(
    storage: Storage,
    output_path: str,
    output: Any,
) -> None:
    storage.put(output_path, pickle.dumps(output))


def _write_json_output(
    storage: Storage,
    output_path: str,
    output: Any,
) -> None:
    storage.put(output_path, json.dumps(output).encode(_DEFAULT_ENCODING))


_serialization_function_mapping = {
    SerializationType.JSON.value: _write_json_output,
    SerializationType.PICKLE.value: _write_pickle_output,
    SerializationType.STRING.value: _write_string_output,
}


def run(spec: MigrationSpec) -> None:
    """
    Executes a artifact migration.
    """
    print("Job Spec: \n{}".format(spec.json()))

    storage = parse_storage(spec.storage_config)
    artifact_metadata = {}

    if spec.artifact_type == "table":
        # Luckily, the serialization logic for table remains the same, so no need to overwrite the content file.
        artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY] = ArtifactType.TABLE.value
        artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.TABLE.value
    elif spec.artifact_type == "float":
        # Luckily, the serialization logic for float remains the same, so no need to overwrite the content file.
        artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY] = ArtifactType.NUMERIC.value
        artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.JSON.value
    elif spec.artifact_type == "boolean":
        artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY] = ArtifactType.BOOL.value
        artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.JSON.value

        artifact_content = bool(storage.get(spec.content_path))
        # The serialization logic for bool is different, so we need to overwrite the content file.
        storage.put(spec.content_path, json.dumps(artifact_content).encode(_DEFAULT_ENCODING))
    elif spec.artifact_type == "json":
        content = json.loads(storage.get(spec.content_path).decode(_DEFAULT_ENCODING))
        # We need to call infer_artifact_type to know its actual type.
        new_artifact_type = infer_artifact_type(content)

        if new_artifact_type == ArtifactType.JSON or new_artifact_type == ArtifactType.STRING:
            artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.STRING.value
        elif new_artifact_type == ArtifactType.BOOL or new_artifact_type == ArtifactType.NUMERIC:
            artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.JSON.value
        elif new_artifact_type == ArtifactType.PICKLABLE:
            artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.PICKLE.value
        elif new_artifact_type == ArtifactType.DICT or new_artifact_type == ArtifactType.TUPLE:
            try:
                json.dumps(content)
                artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.JSON.value
            except:
                artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY] = SerializationType.PICKLE.value
        else:
            raise Exception("Unexpected artifact type %s" % new_artifact_type)

        # The serialization logic might have changed, so we overwrite the content file.
        _serialization_function_mapping[artifact_metadata[_METADATA_SERIALIZATION_TYPE_KEY]](
            storage,
            spec.content_path,
            content,
        )

        artifact_metadata[_METADATA_ARTIFACT_TYPE_KEY] = new_artifact_type.value

    # We always want to update the metadata map to contain the artifact type and serialization type.
    storage.put(spec.metadata_path, json.dumps(artifact_metadata).encode(_DEFAULT_ENCODING))



import os
import sys
import traceback
from typing import Dict, List, Tuple, Union

from aqueduct_executor.operators.airflow import spec
from aqueduct_executor.operators.connectors.data import spec as conn_spec
from aqueduct_executor.operators.function_executor import spec as func_spec
from aqueduct_executor.operators.param_executor import spec as param_spec
from aqueduct_executor.operators.system_metric_executor import spec as system_metric_spec
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.execution import ExecutionState, Logs
from aqueduct_executor.operators.utils.storage import parse
from jinja2 import Environment, FileSystemLoader

TaskSpec = Union[
    conn_spec.ExtractSpec,
    conn_spec.LoadSpec,
    func_spec.FunctionSpec,
    param_spec.ParamSpec,
    system_metric_spec.SystemMetricSpec,
]


class Task:
    def __init__(self, task_id: str, spec: TaskSpec, alias: str):
        self.id = task_id
        self.spec = spec
        self.alias = alias


def run(spec: spec.CompileAirflowSpec) -> None:
    """
    Executes a compile airflow operator.
    """
    print("Started %s job: %s" % (spec.type, spec.name))

    storage = parse.parse_storage(spec.storage_config)
    exec_state = ExecutionState(user_logs=Logs())

    try:
        dag_file = compile(spec)
        utils.write_compile_airflow_output(storage, spec.output_content_path, dag_file)
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
    except Exception:
        traceback.print_exc()
        utils.write_exec_state(storage, spec.metadata_path, exec_state)
        sys.exit(1)


def compile(spec: spec.CompileAirflowSpec) -> bytes:
    """
    Takes a CompileAirflowSpec and generates an Airflow DAG specification Python file.
    It returns the DAG file.
    """

    schedule = None
    if spec.cron_schedule:
        schedule = spec.cron_schedule

    # Init Airflow tasks
    tasks = []
    task_to_alias = {}
    i = 0
    for task_id, task_spec in spec.task_specs.items():
        alias = "t{}".format(i)
        t = Task(task_id, task_spec, alias)
        tasks.append(t)
        i += 1

        task_to_alias[task_id] = alias

    edges = flatten_task_edges(spec.task_edges)

    home = os.environ.get("HOME")
    path = os.path.join(str(home), ".aqueduct", "server/bin")
    env = Environment(loader=FileSystemLoader(path))

    template = env.get_template("dag.template")
    r = template.render(
        workflow_dag_id=spec.workflow_dag_id,
        dag_id=spec.dag_id,
        schedule=schedule,
        tasks=tasks,
        edges=edges,
        task_to_alias=task_to_alias,
    )

    return str.encode(r)


def flatten_task_edges(edges: Dict[str, List[str]]) -> List[Tuple[str, str]]:
    """
    Takes a map of tasks to a list of tasks representing edges and flattens
    it into a list of tuples.
    """
    pairs = []

    for src, dests in edges.items():
        for dest in dests:
            pairs.append((src, dest))

    return pairs
